from __future__ import annotations

import pytest

from gotta.plugins import gdocs, gdrive, gsheets
from gotta.providers import google


def test_parse_doc_ref_accepts_document_url_and_raw_id() -> None:
    assert google.parse_doc_ref("doc-123") == (
        "doc-123",
        "https://docs.google.com/document/d/doc-123/edit",
    )
    assert google.parse_doc_ref("https://docs.google.com/document/d/doc-123/edit") == (
        "doc-123",
        "https://docs.google.com/document/d/doc-123/edit",
    )


def test_parse_doc_ref_rejects_non_document_urls() -> None:
    with pytest.raises(google.GoogleError, match="could not parse Google Docs"):
        google.parse_doc_ref("https://drive.google.com/file/d/drive-123/view")


def test_parse_drive_ref_accepts_drive_and_docs_urls() -> None:
    assert google.parse_drive_ref("drive-123") == (
        "drive-123",
        "https://drive.google.com/open?id=drive-123",
    )
    assert google.parse_drive_ref("https://drive.google.com/file/d/drive-123/view") == (
        "drive-123",
        "https://drive.google.com/file/d/drive-123/view",
    )
    assert google.parse_drive_ref("https://drive.google.com/open?id=drive-123") == (
        "drive-123",
        "https://drive.google.com/open?id=drive-123",
    )
    assert google.parse_drive_ref("https://docs.google.com/spreadsheets/d/sheet-123/edit") == (
        "sheet-123",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit",
    )


def test_parse_sheet_ref_accepts_sheet_url_and_raw_id() -> None:
    assert google.parse_sheet_ref("sheet-123") == (
        "sheet-123",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit",
    )
    assert google.parse_sheet_ref("https://docs.google.com/spreadsheets/d/sheet-123/edit") == (
        "sheet-123",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit",
    )


def test_parse_drive_ref_rejects_unstructured_input() -> None:
    with pytest.raises(google.GoogleError, match="could not parse Google Drive"):
        google.parse_drive_ref("not a valid drive ref")


def test_google_status_payload_reports_missing_session(monkeypatch, tmp_path) -> None:
    monkeypatch.setattr(google, "TOKEN_FILE", tmp_path / "oauth.json")
    monkeypatch.setattr(google, "CONFIG_FILE", tmp_path / "gotta.toml")
    monkeypatch.setattr(google, "load_google_config_env", lambda: {})

    payload = google.google_status_payload()

    assert payload["sessionStatus"] == "missing"
    assert not payload["credentialsConfigured"]


def test_google_status_payload_prefers_native_refresh_before_auth(monkeypatch) -> None:
    monkeypatch.setattr(google, "google_credentials_present", lambda: True)
    monkeypatch.setattr(google, "load_google_config_env", lambda: {})
    monkeypatch.setattr(
        google,
        "load_cached_oauth_state",
        lambda: {
            "access_token": "expired-token",
            "refresh_token": "refresh-token",
            "expires_at": 0.0,
        },
    )

    payload = google.google_status_payload()

    assert payload["sessionStatus"] == "expired"
    assert (
        payload["nextStep"]
        == "rerun a native Google command first; gotta will usually refresh automatically "
        "when a refresh token is present. Re-run auth only if refresh does not recover "
        "the session"
    )


def test_ensure_google_session_distinguishes_missing_session_from_missing_credentials(
    monkeypatch,
) -> None:
    monkeypatch.setattr(google, "load_cached_oauth_state", lambda: None)
    monkeypatch.setattr(google, "google_credentials_present", lambda: True)

    with pytest.raises(google.GoogleError, match="no cached Google OAuth session"):
        google.ensure_google_session(
            allow_bootstrap=False,
            interactive_ok=False,
            auth_command="gdocs",
        )


def test_google_missing_credentials_message_points_to_gotta_config(monkeypatch) -> None:
    monkeypatch.delenv("GOTTA_GOOGLE_OAUTH_CLIENT_ID", raising=False)
    monkeypatch.delenv("GOTTA_GOOGLE_OAUTH_CLIENT_SECRET", raising=False)
    monkeypatch.setattr(google, "load_google_config_env", lambda: {})

    with pytest.raises(google.GoogleError, match=r"\[providers\.google\.env\].*gotta\.toml"):
        google.load_oauth_runtime_config()


def test_google_doc_and_drive_get_default_to_markdown() -> None:
    assert gdocs.build_parser().parse_args(["get", "doc-123"]).output == "markdown"
    assert gdrive.build_parser().parse_args(["get", "file-123"]).output == "markdown"
    assert gsheets.build_parser().parse_args(["get", "sheet-123"]).output == "markdown"


def test_google_status_defaults_to_summary() -> None:
    assert gdocs.build_parser().parse_args(["status"]).output == "summary"
    assert gdocs.build_parser().parse_args(["status", "--output", "summary"]).output == "summary"
    assert gdrive.build_parser().parse_args(["status"]).output == "summary"
    assert gdrive.build_parser().parse_args(["status", "--output", "summary"]).output == "summary"
    assert gsheets.build_parser().parse_args(["status"]).output == "summary"
    assert gsheets.build_parser().parse_args(["status", "--output", "summary"]).output == "summary"


def test_gdocs_search_reports_gdocs_as_source(monkeypatch) -> None:
    monkeypatch.setattr(gdocs, "ensure_google_session", lambda **kwargs: {"access_token": "token"})
    monkeypatch.setattr(
        gdocs,
        "drive_search_files",
        lambda *args, **kwargs: [
            {
                "id": "doc-123",
                "name": "Doc",
                "webViewLink": "https://docs.google.com/document/d/doc-123/edit",
            }
        ],
    )

    payload = gdocs.search_documents("token", "needle", mode="title", limit=5)

    assert payload["source"] == "gdocs"


def test_gdocs_search_markdown_uses_docs_specific_heading() -> None:
    rendered = gdocs.render_search_markdown(
        {
            "query": "needle",
            "mode": "auto",
            "resultCount": 1,
            "results": [
                {
                    "title": "Doc",
                    "url": "https://docs.google.com/document/d/doc-123/edit",
                    "matchedBy": ["title"],
                }
            ],
        }
    )

    assert "### Google Docs Search: needle" in rendered
    assert "- _Surface_: `gdocs`" in rendered


def test_document_meta_merges_drive_created_and_modified(monkeypatch) -> None:
    monkeypatch.setattr(
        gdocs,
        "google_json",
        lambda url, access_token: {
            "documentId": "doc-123",
            "title": "Doc",
            "revisionId": "rev-1",
        },
    )
    monkeypatch.setattr(
        gdocs,
        "drive_file_meta",
        lambda access_token, doc_id, fields: {
            "createdTime": "2026-01-08T17:19:19.520Z",
            "modifiedTime": "2026-01-08T18:49:54.309Z",
            "webViewLink": "https://docs.google.com/document/d/doc-123/edit",
        },
    )

    payload = gdocs.document_meta("token", "doc-123")

    assert payload["createdTime"] == "2026-01-08T17:19:19.520Z"
    assert payload["modifiedTime"] == "2026-01-08T18:49:54.309Z"


def test_gdocs_search_markdown_includes_created_and_updated() -> None:
    rendered = gdocs.render_search_markdown(
        {
            "query": "needle",
            "mode": "auto",
            "resultCount": 1,
            "results": [
                {
                    "title": "Doc",
                    "url": "https://docs.google.com/document/d/doc-123/edit",
                    "matchedBy": ["title"],
                    "createdTime": "2026-01-08T17:19:19.520Z",
                    "modifiedTime": "2026-01-08T18:49:54.309Z",
                }
            ],
        }
    )

    assert "- Created: 2026-01-08T17:19:19.520Z" in rendered
    assert "- Updated: 2026-01-08T18:49:54.309Z" in rendered
    assert "created `2026-01-08T17:19:19.520Z`" in rendered
    assert "modified `2026-01-08T18:49:54.309Z`" in rendered


def test_gdocs_get_markdown_includes_metadata_header(monkeypatch, capsys) -> None:
    monkeypatch.setattr(gdocs, "is_interactive", lambda: False)
    monkeypatch.setattr(
        gdocs,
        "ensure_google_session",
        lambda **kwargs: {"access_token": "token"},
    )
    monkeypatch.setattr(gdocs, "parse_doc_ref", lambda ref: ("doc-123", ""))
    monkeypatch.setattr(
        gdocs,
        "document_meta",
        lambda access_token, doc_id: {
            "url": "https://docs.google.com/document/d/doc-123/edit",
            "createdTime": "2026-01-08T17:19:19.520Z",
            "modifiedTime": "2026-01-08T18:49:54.309Z",
            "revisionId": "rev-1",
            "owners": [{"displayName": "Alex Example"}],
        },
    )
    monkeypatch.setattr(gdocs, "drive_export", lambda access_token, doc_id, _mime: b"# Body\n")

    result = gdocs.main(["get", "doc-123", "--output", "markdown"])

    captured = capsys.readouterr()
    assert result == 0
    assert "- Created: 2026-01-08T17:19:19.520Z" in captured.out
    assert "- Updated: 2026-01-08T18:49:54.309Z" in captured.out
    assert "- Revision: rev-1" in captured.out
    assert "# Body" in captured.out


def test_gsheets_search_reports_gsheets_as_source(monkeypatch) -> None:
    monkeypatch.setattr(
        gsheets,
        "drive_search_files",
        lambda *args, **kwargs: [
            {
                "id": "sheet-123",
                "name": "Sheet",
                "webViewLink": "https://docs.google.com/spreadsheets/d/sheet-123/edit",
                "mimeType": google.GOOGLE_SHEET_MIME,
            }
        ],
    )

    payload = gsheets.search_spreadsheets("token", "needle", mode="title", limit=5)

    assert payload["source"] == "gsheets"


def test_gsheets_search_markdown_uses_sheets_specific_heading() -> None:
    rendered = gsheets.render_search_markdown(
        {
            "query": "needle",
            "mode": "auto",
            "resultCount": 1,
            "results": [
                {
                    "title": "Sheet",
                    "url": "https://docs.google.com/spreadsheets/d/sheet-123/edit",
                    "matchedBy": ["title"],
                }
            ],
        }
    )

    assert "### Google Sheets Search: needle" in rendered
    assert "- _Surface_: `gsheets`" in rendered


def test_gsheets_search_markdown_includes_time_envelope() -> None:
    rendered = gsheets.render_search_markdown(
        {
            "query": "needle",
            "mode": "auto",
            "resultCount": 1,
            "results": [
                {
                    "title": "Sheet",
                    "url": "https://docs.google.com/spreadsheets/d/sheet-123/edit",
                    "matchedBy": ["title"],
                    "createdTime": "2026-01-08T17:19:19.520Z",
                    "modifiedTime": "2026-01-08T18:49:54.309Z",
                }
            ],
        }
    )

    assert "- Created: 2026-01-08T17:19:19.520Z" in rendered
    assert "- Updated: 2026-01-08T18:49:54.309Z" in rendered


def test_gsheets_summary_includes_created_and_modified() -> None:
    rendered = gsheets.spreadsheet_summary(
        {
            "spreadsheetId": "sheet-123",
            "url": "https://docs.google.com/spreadsheets/d/sheet-123/edit",
            "createdTime": "2026-01-08T17:19:19.520Z",
            "modifiedTime": "2026-01-08T18:49:54.309Z",
            "properties": {"title": "Sheet"},
            "sheets": [{"properties": {"title": "Tab 1"}}],
        }
    )

    assert "- **Created:** 2026-01-08T17:19:19.520Z" in rendered
    assert "- **Modified:** 2026-01-08T18:49:54.309Z" in rendered


def test_gdrive_summary_includes_created_and_modified() -> None:
    rendered = gdrive.drive_file_summary(
        {
            "id": "file-123",
            "name": "File",
            "createdTime": "2026-01-08T17:19:19.520Z",
            "modifiedTime": "2026-01-08T18:49:54.309Z",
        }
    )

    assert "- **Created:** 2026-01-08T17:19:19.520Z" in rendered
    assert "- **Modified:** 2026-01-08T18:49:54.309Z" in rendered


def test_gdrive_search_markdown_includes_time_envelope() -> None:
    rendered = gdrive.render_search_markdown(
        {
            "query": "needle",
            "mode": "auto",
            "resultCount": 1,
            "results": [
                {
                    "title": "File",
                    "url": "https://drive.google.com/file/d/file-123/view",
                    "matchedBy": ["title"],
                    "createdTime": "2026-01-08T17:19:19.520Z",
                    "modifiedTime": "2026-01-08T18:49:54.309Z",
                }
            ],
        }
    )

    assert "- Created: 2026-01-08T17:19:19.520Z" in rendered
    assert "- Updated: 2026-01-08T18:49:54.309Z" in rendered


def test_gdocs_auth_prefers_reuse_or_refresh(monkeypatch, capsys) -> None:
    monkeypatch.setattr(gdocs, "is_interactive", lambda: True)
    monkeypatch.setattr(
        gdocs,
        "ensure_google_session",
        lambda **kwargs: {"access_token": "token", "expires_at": 123.0},
    )
    monkeypatch.setattr(
        gdocs,
        "run_oauth_bootstrap",
        lambda **kwargs: pytest.fail("full bootstrap should not run by default"),
    )

    result = gdocs.main(["auth"])

    captured = capsys.readouterr()
    assert result == 0
    assert '"authenticated": true' in captured.out
    assert '"expires_at": 123.0' in captured.out
    assert '"token_file":' in captured.out


def test_gdocs_auth_full_forces_browser_bootstrap(monkeypatch, capsys) -> None:
    monkeypatch.setattr(gdocs, "is_interactive", lambda: True)
    monkeypatch.setattr(
        gdocs,
        "ensure_google_session",
        lambda **kwargs: pytest.fail("default refresh path should be bypassed with --full"),
    )
    monkeypatch.setattr(
        gdocs,
        "run_oauth_bootstrap",
        lambda **kwargs: {"access_token": "token", "expires_at": 456.0},
    )

    result = gdocs.main(["auth", "--full"])

    captured = capsys.readouterr()
    assert result == 0
    assert '"authenticated": true' in captured.out
    assert '"expires_at": 456.0' in captured.out
    assert '"token_file":' in captured.out


def test_gdrive_auth_prefers_reuse_or_refresh(monkeypatch, capsys) -> None:
    monkeypatch.setattr(gdrive, "is_interactive", lambda: True)
    monkeypatch.setattr(
        gdrive,
        "ensure_google_session",
        lambda **kwargs: {"access_token": "token", "expires_at": 123.0},
    )
    monkeypatch.setattr(
        gdrive,
        "run_oauth_bootstrap",
        lambda **kwargs: pytest.fail("full bootstrap should not run by default"),
    )

    result = gdrive.main(["auth"])

    captured = capsys.readouterr()
    assert result == 0
    assert '"authenticated": true' in captured.out
    assert '"expires_at": 123.0' in captured.out
    assert '"token_file":' in captured.out


def test_gdrive_auth_full_forces_browser_bootstrap(monkeypatch, capsys) -> None:
    monkeypatch.setattr(gdrive, "is_interactive", lambda: True)
    monkeypatch.setattr(
        gdrive,
        "ensure_google_session",
        lambda **kwargs: pytest.fail("default refresh path should be bypassed with --full"),
    )
    monkeypatch.setattr(
        gdrive,
        "run_oauth_bootstrap",
        lambda **kwargs: {"access_token": "token", "expires_at": 456.0},
    )

    result = gdrive.main(["auth", "--full"])

    captured = capsys.readouterr()
    assert result == 0
    assert '"authenticated": true' in captured.out
    assert '"expires_at": 456.0' in captured.out
    assert '"token_file":' in captured.out



from __future__ import annotations

import io
import json
from pathlib import Path

import pytest

from gotta.actors import ACTOR_CALLEE_ENV, ACTOR_SPEAKER_ENV
from gotta import content, dispatch
from gotta import leads
from gotta import main as cli
from gotta.peer import PEER_SESSION_ACTOR_ENV
from gotta import session as sessionlib
from gotta import todo as session_todo
from gotta.notes import peer_notes_surface_path
from gotta.plugins import goal
from gotta.plugins import notes
from gotta.plugins import peer
from gotta.plugins import session
from gotta.plugins import want


@pytest.fixture(autouse=True)
def local_session_registry(tmp_path: Path, monkeypatch) -> None:
    registry = tmp_path / "sessions"
    monkeypatch.setattr(content, "DEFAULT_SESSION_ROOT", registry)
    monkeypatch.setattr(cli, "DEFAULT_SESSION_ROOT", registry)
    monkeypatch.delenv(content.SESSION_ENV, raising=False)
    monkeypatch.delenv(content.CONTENT_ENV, raising=False)
    monkeypatch.delenv(content.SESSION_REPO_ENV, raising=False)
    monkeypatch.delenv(ACTOR_SPEAKER_ENV, raising=False)
    monkeypatch.delenv(ACTOR_CALLEE_ENV, raising=False)
    monkeypatch.delenv(PEER_SESSION_ACTOR_ENV, raising=False)


def make_dirs(root: Path) -> content.ResolvedDirs:
    dirs = content.ResolvedDirs(
        session_dir=root,
        content_dir=root / "content",
    )
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    return dirs


def initialize_session(root: Path) -> content.ResolvedDirs:
    dirs = make_dirs(root)
    content.write_state_env(dirs)
    dirs.session_dir.joinpath("bin").mkdir(parents=True, exist_ok=True)
    return dirs


def _init_session(root: Path, capsys) -> None:
    assert session.main(["init", "--session", str(root)]) == 0
    assert capsys.readouterr().out.strip() == str(root.resolve())


def _configure_peers(root: Path, capsys, *actors: str) -> str:
    assert peer.main(["with", *actors, "--session", str(root)]) == 0
    return capsys.readouterr().out


@pytest.mark.parametrize("command", ["show", "doctor"])
def test_session_commands_require_bootstrap(
    tmp_path: Path, monkeypatch, capsys, command: str
) -> None:
    default_root = tmp_path / "default"
    monkeypatch.setattr(content, "DEFAULT_SESSION_ROOT", default_root)
    monkeypatch.chdir(tmp_path)

    with pytest.raises(SystemExit) as excinfo:
        session.main([command])

    assert excinfo.value.code == 2
    assert "start or bind a session first" in capsys.readouterr().err


def test_session_init_bootstraps_state_env_and_bin(
    tmp_path: Path,
    capsys,
) -> None:
    root = tmp_path / "session-root"

    assert session.main(["init", "--session", str(root)]) == 0

    assert capsys.readouterr().out.strip() == str(root.resolve())
    assert (root / "state" / "env").exists()
    assert (root / "state").is_dir()
    assert (root / "bin").is_dir()


def test_session_init_scaffolds_surface_and_preserves_context_state(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=root, content_dir=root / "content")
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content.write_session_state(
        dirs,
        {
            content.CONTEXT_ID_ENV: "ctx-123",
            content.CONTEXT_SOURCE_ENV: "test",
        },
    )

    _init_session(root, capsys)

    state = content.load_state_env_at_root(root)
    assert state[content.CONTEXT_ID_ENV] == "ctx-123"
    assert state[content.CONTEXT_SOURCE_ENV] == "test"
    assert state[content.WORK_INITIALIZED_ENV] == "1"
    for name in ("WANT.md", "GOAL.md", "TODO.md", "LOGS.md", "OOPS.md"):
        assert (root / name).exists()
    assert not (root / "WORK.md").exists()
    assert "_empty_" in (root / "WANT.md").read_text(encoding="utf-8")
    assert "Mission seed" not in (root / "WANT.md").read_text(encoding="utf-8")


def test_session_init_is_idempotent_and_preserves_rewritten_charters(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    want_text = "# Want\n\nStay narrow.\n"
    goal_text = "# Goal\n\nCollect two anchors.\n"
    (root / "WANT.md").write_text(want_text, encoding="utf-8")
    (root / "GOAL.md").write_text(goal_text, encoding="utf-8")
    initial_todo_count = len(session_todo.todo_items(root))

    _init_session(root, capsys)

    assert (root / "WANT.md").read_text(encoding="utf-8") == want_text
    assert (root / "GOAL.md").read_text(encoding="utf-8") == goal_text
    assert len(session_todo.todo_items(root)) == initial_todo_count


def test_want_and_goal_support_show_and_session_relative_from_file(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"
    charter_dir = root / "charters"
    charter_dir.mkdir(parents=True)
    want_source = charter_dir / "want.txt"
    goal_source = charter_dir / "goal.txt"
    want_source.write_text("# Want\n\nTrack retry ownership.\n", encoding="utf-8")
    goal_source.write_text("# Goal\n\nMaterialize one strong source first.\n", encoding="utf-8")

    _init_session(root, capsys)

    assert want.main(["--session", str(root), "--from-file", "charters/want.txt"]) == 0
    capsys.readouterr()
    assert goal.main(["--session", str(root), "--from-file", "charters/goal.txt"]) == 0
    capsys.readouterr()

    assert want.main(["--session", str(root)]) == 0
    want_output = capsys.readouterr().out
    assert want_output == "# Want\n\nTrack retry ownership.\n"

    assert goal.main(["--session", str(root)]) == 0
    goal_output = capsys.readouterr().out
    assert goal_output == "# Goal\n\nMaterialize one strong source first.\n"


def test_want_and_goal_support_explicit_stdin_rewrites(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)

    monkeypatch.setattr(sessionlib.sys, "stdin", io.StringIO("# Want\n\nStay focused.\n"))
    assert want.main(["--session", str(root), "--stdin"]) == 0
    capsys.readouterr()

    monkeypatch.setattr(sessionlib.sys, "stdin", io.StringIO("# Goal\n\nLand one anchor.\n"))
    assert goal.main(["--session", str(root), "--stdin"]) == 0
    capsys.readouterr()

    assert (root / "WANT.md").read_text(encoding="utf-8") == "# Want\n\nStay focused.\n"
    assert (root / "GOAL.md").read_text(encoding="utf-8") == "# Goal\n\nLand one anchor.\n"


def test_want_and_goal_reject_inline_positional_text(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)

    with pytest.raises(SystemExit) as want_exc:
        want.main(["inline text", "--session", str(root)])
    assert want_exc.value.code == 2

    with pytest.raises(SystemExit) as goal_exc:
        goal.main(["inline text", "--session", str(root)])
    assert goal_exc.value.code == 2


def test_want_and_goal_can_target_linked_peer_sessions(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    monkeypatch.setenv(content.SESSION_ENV, str(root))
    monkeypatch.setenv(content.CONTENT_ENV, str(root / "content"))
    peer_root = sessionlib._peer_session_dir(root, "claude")
    charter_dir = peer_root / "charters"
    charter_dir.mkdir(parents=True, exist_ok=True)
    (charter_dir / "want.txt").write_text("Peer-specific intent\n", encoding="utf-8")
    (charter_dir / "goal.txt").write_text("Peer-specific goal\n", encoding="utf-8")

    assert want.main(["--session", "peers/claude", "--from-file", "charters/want.txt"]) == 0
    capsys.readouterr()
    assert goal.main(["--session", "peers/claude", "--from-file", "charters/goal.txt"]) == 0
    capsys.readouterr()

    assert (peer_root / "WANT.md").read_text(encoding="utf-8") == "Peer-specific intent\n"
    assert (peer_root / "GOAL.md").read_text(encoding="utf-8") == "Peer-specific goal\n"


def test_peer_with_configures_linked_peer_surfaces_without_launching(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    output = _configure_peers(root, capsys, "Claude", "Codex")

    assert "configured Claude workspace" in output
    assert "configured Codex workspace" in output
    assert "gotta want --session peers/claude --stdin" in output
    assert "gotta goal --session peers/claude --stdin" in output
    for peer_name in ("claude", "codex"):
        peer_root = sessionlib._peer_session_dir(root, peer_name)
        assert peer_root.exists()
        assert (root / "peers" / peer_name).resolve() == peer_root.resolve()
        assert (root / "bin" / peer_name).is_file()
        assert (peer_root / "README.md").is_file()
        assert (peer_root / "TODO.md").is_file()
        assert sessionlib._read_peer_state(root, peer_name)["status"] == "configured"


def test_peer_launch_blockers_emit_native_peer_charter_commands(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")

    blockers = sessionlib._peer_launch_blockers(root, peer_name="claude")

    assert any("gotta want --session peers/claude --stdin" in blocker for blocker in blockers)
    assert any("gotta goal --session peers/claude --stdin" in blocker for blocker in blockers)


def test_peer_launch_blockers_report_linked_peer_paths(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")

    blockers = sessionlib._peer_launch_blockers(root, peer_name="claude")

    assert any(str(root / "peers" / "claude" / "WANT.md") in blocker for blocker in blockers)
    assert any(str(root / "peers" / "claude" / "GOAL.md") in blocker for blocker in blockers)
    assert not any(
        str(sessionlib._peer_session_dir(root, "claude")) in blocker for blocker in blockers
    )


def test_peer_status_empty_guidance_points_to_peer_with(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)

    assert peer.main(["status", "--session", str(root)]) == 0
    output = capsys.readouterr().out
    assert "no peers configured for this session" in output
    assert "gotta peer with Claude" in output


def test_notes_show_empty_guidance_points_to_peer_with(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)

    assert notes.main(["show", "--session", str(root)]) == 0
    output = capsys.readouterr().out
    assert "no peers configured for this session" in output
    assert "gotta peer with Claude" in output


def test_notes_show_unconfigured_peer_fails_without_materializing_surface(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)

    with pytest.raises(SystemExit) as excinfo:
        notes.main(["show", "claude", "--session", str(root)])

    assert "claude is not configured for this session" in str(excinfo.value)
    assert not (root / "peers" / "claude").exists()
    assert not (root / "state" / "peers" / "claude").exists()


def test_peer_complete_becomes_terminal_immediately_and_survives_heartbeat(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(root, "claude", {"status": "active"})

    assert (
        peer.main(
            [
                "complete",
                "claude",
                "--session",
                str(root),
                "--summary",
                "ready for review",
            ]
        )
        == 0
    )
    capsys.readouterr()
    payload = sessionlib._peer_status_payload(root, "claude")
    assert payload["status"] == "completed"
    assert payload["requested_pending"] is False
    assert payload["summary"] == "ready for review"

    assert peer.main(["heartbeat", "claude", "--session", str(root)]) == 0
    capsys.readouterr()
    assert sessionlib._peer_status_payload(root, "claude")["status"] == "completed"


def test_peer_signoff_becomes_terminal_immediately_and_survives_heartbeat(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(root, "claude", {"status": "active"})

    assert (
        peer.main(
            [
                "signoff",
                "claude",
                "--session",
                str(root),
                "--summary",
                "accepted by operator",
            ]
        )
        == 0
    )
    capsys.readouterr()
    payload = sessionlib._peer_status_payload(root, "claude")
    assert payload["status"] == "signed_off"
    assert payload["requested_pending"] is False
    assert payload["signoff_summary"] == "accepted by operator"

    assert peer.main(["heartbeat", "claude", "--session", str(root)]) == 0
    capsys.readouterr()
    assert sessionlib._peer_status_payload(root, "claude")["status"] == "signed_off"


def test_peer_status_treats_signoff_timestamp_as_authoritative(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(
        root,
        "claude",
        {
            "status": "active",
            "heartbeat_at": "2026-03-17T00:00:00Z",
            "signoff_at": "2026-03-17T00:01:00Z",
            "signoff_summary": "peer closed itself cleanly",
        },
    )

    payload = sessionlib._peer_status_payload(root, "claude")

    assert payload["status"] == "signed_off"
    assert payload["still_running"] is False
    assert payload["signoff_summary"] == "peer closed itself cleanly"


def test_peer_fail_stays_pending_while_live_and_notes_render_stop_warning(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(root, "claude", {"status": "active"})

    assert (
        peer.main(
            [
                "fail",
                "claude",
                "--session",
                str(root),
                "--summary",
                "operator stopped this run",
            ]
        )
        == 0
    )
    output = capsys.readouterr().out
    payload = sessionlib._peer_status_payload(root, "claude")

    assert "authoritative status stays active" in output
    assert payload["status"] == "active"
    assert payload["requested_pending"] is True
    assert payload["requested_status"] == "failed"
    notes_text = peer_notes_surface_path(root, "claude").read_text(encoding="utf-8")
    assert "Supervisor requested `failed` (operator stopped this run)." in notes_text
    assert "pending_disposition: failed: operator stopped this run" in notes_text


def test_peer_stop_stays_pending_while_live_and_notes_render_graceful_warning(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(root, "claude", {"status": "active"})

    assert (
        peer.main(
            [
                "stop",
                "claude",
                "--session",
                str(root),
                "--summary",
                "finish the current wave and close out",
            ]
        )
        == 0
    )
    output = capsys.readouterr().out
    payload = sessionlib._peer_status_payload(root, "claude")

    assert "recorded stop request for claude" in output
    assert payload["status"] == "active"
    assert payload["requested_pending"] is True
    assert payload["requested_status"] == "signed_off"
    assert payload["requested_mode"] == "stop"
    assert payload["requested_label"] == "stop"
    notes_text = peer_notes_surface_path(root, "claude").read_text(encoding="utf-8")
    assert (
        "Supervisor requested a graceful stop (finish the current wave and close out)."
        in notes_text
    )
    assert "pending_disposition: stop: finish the current wave and close out" in notes_text


def test_notes_projection_skips_supervisor_warning_for_nonfailed_pending_requests(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(
        root,
        "claude",
        {
            "status": "active",
            "requested_status": "signed_off",
            "requested_summary": "looked done from the operator side",
        },
    )
    sessionlib._sync_peer_projection_surfaces(root, "claude")

    notes_text = peer_notes_surface_path(root, "claude").read_text(encoding="utf-8")
    assert "Supervisor requested `failed`" not in notes_text
    assert "pending_disposition: signed off: looked done from the operator side" in notes_text


def test_peer_with_reconciles_preexisting_shared_log_file(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    peer_root = sessionlib._peer_session_dir(root, "claude")
    peer_root.mkdir(parents=True, exist_ok=True)
    (peer_root / "LOGS.md").write_text("stale log placeholder\n", encoding="utf-8")
    peer_state_dir = peer_root / "state"
    peer_state_dir.mkdir(parents=True, exist_ok=True)
    (peer_state_dir / "peer.json").write_text(
        json.dumps(
            {
                "peer": "claude",
                "label": "Claude",
                "status": "signed_off",
                "signoff_summary": "stale orphaned peer state",
            }
        )
        + "\n",
        encoding="utf-8",
    )
    (peer_root / "NOTES.md").write_text("# Claude Notes\n\nstale\n", encoding="utf-8")

    assert peer.main(["with", "Claude", "--session", str(root)]) == 0
    capsys.readouterr()

    assert (peer_root / "LOGS.md").is_symlink()
    assert (peer_root / "LOGS.md").resolve() == (root / "LOGS.md").resolve()
    assert sessionlib._peer_status_payload(root, "claude")["status"] == "configured"


def test_peer_status_reports_recent_activity_and_recent_artifacts(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(
        root,
        "claude",
        {
            "status": "signed_off",
            "signoff_at": "2026-03-17T00:03:00Z",
            "signoff_summary": "accepted by operator",
        },
    )
    events_path = sessionlib._peer_events_path(root, "claude")
    events_path.parent.mkdir(parents=True, exist_ok=True)
    events_path.write_text(
        "\n".join(
            [
                json.dumps(
                    {
                        "timestamp": "2026-03-17T00:01:00Z",
                        "peer": "claude",
                        "event": "heartbeat",
                        "detail": "",
                    }
                ),
                json.dumps(
                    {
                        "timestamp": "2026-03-17T00:02:00Z",
                        "peer": "claude",
                        "event": "signed_off",
                        "detail": "accepted by operator",
                    }
                ),
                json.dumps(
                    {
                        "timestamp": "2026-03-17T00:03:00Z",
                        "peer": "claude",
                        "event": "note",
                        "detail": "Wave 2 landed",
                    }
                ),
            ]
        )
        + "\n",
        encoding="utf-8",
    )
    dirs = content.ResolvedDirs(session_dir=root, content_dir=root / "content")
    for index, locator in enumerate(
        [
            "jira:DEMO-6292",
            "jira:DEMO-6361",
            "github:search --type pr edge rmp acme",
            "jira:search RMP",
        ],
        start=1,
    ):
        content.materialize_bytes(
            f"artifact {index}".encode("utf-8"),
            dirs=dirs,
            preferred_name=f"artifact-{index}.md",
            metadata={
                "tool": "gotta",
                "plugin": "jira" if locator.startswith("jira:") else "github",
                "locator": locator,
                "canonical_locator": locator,
                "actor": "claude",
            },
            timestamp=f"2026-03-17T00:0{index}:00.000001Z",
        )

    assert peer.main(["status", "claude", "--session", str(root)]) == 0
    output = capsys.readouterr().out
    assert "artifacts: 4" in output
    assert "recent_activity: 2026-03-17T00:03:00Z Wave 2 landed" in output
    assert "recent_artifacts:" in output
    assert "`jira:search RMP`" in output
    assert "`github:search --type pr edge rmp acme`" in output
    assert "`jira:DEMO-6361`" in output
    assert "(+1 more)" in output
    assert "heartbeat" not in output


def test_peer_status_json_reports_recent_activity_and_last_activity_summary(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session"

    _init_session(root, capsys)
    _configure_peers(root, capsys, "Claude")
    sessionlib._write_peer_state(root, "claude", {"status": "active"})
    events_path = sessionlib._peer_events_path(root, "claude")
    events_path.parent.mkdir(parents=True, exist_ok=True)
    events_path.write_text(
        "\n".join(
            [
                json.dumps(
                    {
                        "timestamp": "2026-03-17T00:00:00Z",
                        "peer": "claude",
                        "event": "heartbeat",
                        "detail": "",
                    }
                ),
                json.dumps(
                    {
                        "timestamp": "2026-03-17T00:01:00Z",
                        "peer": "claude",
                        "event": "note",
                        "detail": "Wave 1 landed",
                    }
                ),
                json.dumps(
                    {
                        "timestamp": "2026-03-17T00:02:00Z",
                        "peer": "claude",
                        "event": "runtime_exit",
                        "detail": "peer process exited with code 0",
                    }
                ),
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    assert peer.main(["status", "claude", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)["claude"]
    assert payload["last_activity_at"] == "2026-03-17T00:02:00Z"
    assert payload["last_activity_summary"] == "runtime exit: peer process exited with code 0"
    assert [item["event"] for item in payload["recent_activity"]] == ["runtime_exit", "note"]
    assert payload["artifact_count"] == 0


def test_work_plugin_is_removed_without_redirect_shims(capsys) -> None:
    assert cli.main(["work"]) == 2
    err = capsys.readouterr().err
    assert "unknown gotta plugin: work" in err
    assert "peer with" not in err

    assert cli.main(["work", "on", "legacy mission"]) == 2
    err = capsys.readouterr().err
    assert "unknown gotta plugin: work" in err


def test_session_show_works_from_initialized_session_root(tmp_path: Path, monkeypatch, capsys) -> None:
    local_root = tmp_path / "local"
    initialize_session(local_root)

    monkeypatch.chdir(local_root)

    assert session.main(["show"]) == 0
    assert capsys.readouterr().out.strip() == str(local_root.resolve())


@pytest.mark.parametrize(
    "argv",
    [
        ["show"],
        ["doctor"],
        ["manifest", "--output", "json"],
        ["timeline", "--output", "json"],
        ["graph", "--output", "json"],
    ],
)
def test_session_commands_accept_explicit_root_without_ambient_activation(
    tmp_path: Path, monkeypatch, capsys, argv: list[str]
) -> None:
    local_root = tmp_path / "local"
    initialize_session(local_root)

    monkeypatch.chdir(tmp_path)

    assert session.main([argv[0], "--session", str(local_root), *argv[1:]]) == 0

    output = capsys.readouterr().out
    assert output.strip()


def test_session_show_reports_local_state_over_default(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    default_root = tmp_path / "default"
    local_root = tmp_path / "local"
    initialize_session(default_root)
    initialize_session(local_root)

    monkeypatch.setattr(content, "DEFAULT_SESSION_ROOT", default_root)
    monkeypatch.chdir(local_root)

    assert session.main([]) == 0
    assert capsys.readouterr().out.strip() == str(local_root.resolve())


def test_session_analyze_writes_summary_and_graph(tmp_path: Path, monkeypatch, capsys) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    result = content.materialize_bytes(
        b"hello world",
        dirs=dirs,
        preferred_name="demo.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": "demo",
            "canonical_locator": "demo",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"hello again",
        dirs=dirs,
        preferred_name="demo.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": "demo-edit",
            "canonical_locator": "demo",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )
    monkeypatch.chdir(local_root)

    assert session.main(["analyze", "--session", str(local_root)]) == 0
    summary = json.loads(capsys.readouterr().out)
    assert summary["sessionDir"] == str(local_root.resolve())
    assert summary["contentCount"] == 2
    assert summary["revisionEdgeCount"] == 1
    assert summary["semanticNodeCount"] >= 2
    assert Path(summary["graphMermaidPath"]).read_text(encoding="utf-8").startswith("---")
    graph_markdown = Path(summary["graphMermaidMarkdownPath"]).read_text(encoding="utf-8")
    assert graph_markdown.startswith("# gotta session analysis\n\n```mermaid\n")
    assert 'flowchart LR' in graph_markdown
    assert summary["graphMermaidMarkdownPath"].endswith(".mmd.md")
    graph_payload = json.loads(Path(summary["graphJsonPath"]).read_text(encoding="utf-8"))
    semantic_payload = json.loads(
        Path(summary["semanticGraphJsonPath"]).read_text(encoding="utf-8")
    )
    semantic_graph_markdown = Path(summary["semanticGraphMermaidMarkdownPath"]).read_text(
        encoding="utf-8"
    )
    assert semantic_graph_markdown.startswith("# gotta semantic graph\n\n```mermaid\n")
    assert summary["semanticGraphMermaidMarkdownPath"].endswith(".mmd.md")
    assert graph_payload["sessionDir"] == str(local_root.resolve())
    assert graph_payload["contentCount"] == 2
    assert semantic_payload["nodeCount"] >= 2
    assert any(edge["to"] != edge["from"] for edge in graph_payload["revisionEdges"])
    assert result.data_path.name == "data"


def test_session_graph_prefers_canonical_locator_for_binding(tmp_path: Path) -> None:
    dirs = initialize_session(tmp_path / "local")
    manifest_path = dirs.content_dir / "manifest.jsonl"
    manifest_path.write_text(
        "\n".join(
            [
                json.dumps(
                    {
                        "locator": "https://example.atlassian.net/browse/PROJ-3960",
                        "canonical_locator": "jira:PROJ-3960",
                        "plugin": "read",
                        "checksum": "abc123",
                        "preferred_name": "PROJ-3960.md",
                    }
                ),
                json.dumps(
                    {
                        "locator": "get https://example.atlassian.net/browse/PROJ-3960 --output markdown",
                        "canonical_locator": "jira:PROJ-3960",
                        "plugin": "jira",
                        "checksum": "abc123",
                        "preferred_name": "PROJ-3960.md",
                    }
                ),
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    payload = session._graph_payload(dirs)

    assert payload["sourceCount"] == 1
    assert payload["contentCount"] == 1
    assert payload["edgeCount"] == 2
    assert payload["sources"] == [
        {
            "locator": "jira:PROJ-3960",
            "followCommand": "gotta read 'jira:PROJ-3960'",
            "contentCount": 1,
            "collision": False,
            "variant": False,
            "variantCount": 0,
            "variants": [],
        }
    ]
    assert payload["content"][0]["contentLocator"] == "content:abc123"
    assert payload["content"][0]["artifactLocator"] == "artifact:PROJ-3960.md@abc123"
    assert payload["content"][0]["followCommand"] == "gotta read 'artifact:PROJ-3960.md@abc123'"


def test_session_empty_graph_and_analyze_make_empty_state_explicit(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    initialize_session(local_root)
    monkeypatch.chdir(tmp_path)

    assert session.main(["graph", "--session", str(local_root)]) == 0
    graph_output = capsys.readouterr().out
    assert "No evidence nodes yet." in graph_output
    assert "gotta read &lt;locator&gt;" in graph_output

    assert session.main(["graph", "--session", str(local_root), "--output", "json"]) == 0
    graph_payload = json.loads(capsys.readouterr().out)
    assert graph_payload["empty"] is True
    assert graph_payload["nextStep"].startswith("No evidence nodes yet.")

    assert session.main(["analyze", "--session", str(local_root), "--stdout"]) == 0
    analyze_output = capsys.readouterr().out
    assert "No evidence nodes yet." in analyze_output
    assert "gotta read &lt;locator&gt;" in analyze_output


def test_session_analyze_treats_multiple_renderings_as_variants_not_collisions(
    tmp_path: Path,
) -> None:
    dirs = initialize_session(tmp_path / "local")
    url = "https://github.com/acme/widgets"
    content.materialize_bytes(
        b"summary view",
        dirs=dirs,
        preferred_name="widgets.summary",
        metadata={
            "tool": "gotta",
            "plugin": "github",
            "subcommand": "default",
            "argv": ["--output", "summary", url],
            "locator": url,
            "canonical_locator": url,
            "content_type": "text/plain",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"markdown view",
        dirs=dirs,
        preferred_name="widgets.md",
        metadata={
            "tool": "gotta",
            "plugin": "github",
            "subcommand": "default",
            "argv": ["--output", "markdown", url],
            "locator": url,
            "canonical_locator": url,
            "content_type": "text/markdown",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )

    graph_payload = session._graph_payload(dirs)
    analysis_payload = session._analysis_payload(dirs)

    assert graph_payload["sources"][0]["collision"] is False
    assert graph_payload["sources"][0]["variant"] is True
    assert graph_payload["sources"][0]["variantCount"] == 2
    assert analysis_payload["collisionCount"] == 0
    assert analysis_payload["variantCount"] == 1
    assert analysis_payload["revisionEdgeCount"] == 0
    assert analysis_payload["sources"][0]["variant"] is True
    assert analysis_payload["sources"][0]["variantCount"] == 2

    mermaid = session._render_analysis_mermaid(analysis_payload)
    assert "renderings: 2" in mermaid


def test_session_analyze_reports_duplicate_materializations_without_variant_drift(
    tmp_path: Path,
) -> None:
    dirs = initialize_session(tmp_path / "local")
    url = "confluence:4373708801"
    content.materialize_bytes(
        b"primary body",
        dirs=dirs,
        preferred_name="4373708801.md",
        metadata={
            "tool": "gotta",
            "plugin": "confluence",
            "locator": "get 4373708801",
            "canonical_locator": url,
            "actor": "primary",
            "content_type": "text/markdown",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"peer body",
        dirs=dirs,
        preferred_name="4373708801.md",
        metadata={
            "tool": "gotta",
            "plugin": "confluence",
            "locator": "get 4373708801",
            "canonical_locator": url,
            "actor": "claude",
            "content_type": "text/markdown",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )

    analysis_payload = session._analysis_payload(dirs)

    assert analysis_payload["collisionCount"] == 0
    assert analysis_payload["variantCount"] == 0
    assert analysis_payload["duplicateMaterializationCount"] == 1
    assert analysis_payload["duplicateMaterializations"] == [url]
    assert analysis_payload["sources"][0]["duplicateMaterialization"] is True
    assert analysis_payload["sources"][0]["contentCount"] == 2

    mermaid = session._render_analysis_mermaid(analysis_payload)
    assert "materializations: 2" in mermaid


def test_session_analyze_extracts_explicit_leads_and_surfaces_gaps(
    tmp_path: Path,
) -> None:
    dirs = initialize_session(tmp_path / "local")
    content.materialize_bytes(
        (
            "# ABC-1\n\n"
            "Depends on ABC-2.\n"
            "Design doc: confluence:12345\n"
            "PR: https://github.com/acme/widgets/pull/7\n"
        ).encode("utf-8"),
        dirs=dirs,
        preferred_name="ABC-1.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-1",
            "canonical_locator": "jira:ABC-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"# ABC-2\n\nDone.\n",
        dirs=dirs,
        preferred_name="ABC-2.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-2",
            "canonical_locator": "jira:ABC-2",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )

    analysis_payload = session._analysis_payload(dirs)
    semantic_payload = session._semantic_payload(dirs)
    analysis_mermaid = session._render_analysis_mermaid(analysis_payload)

    assert analysis_payload["leadSourceCount"] == 3
    assert analysis_payload["materializedLeadSourceCount"] == 1
    assert analysis_payload["unmaterializedLeadSourceCount"] == 2
    assert analysis_payload["leadEdgeCount"] == 3
    assert any(edge["targetLocator"] == "jira:ABC-2" for edge in analysis_payload["leadEdges"])
    assert any(edge["targetLocator"] == "confluence:12345" for edge in analysis_payload["leadEdges"])
    assert any(
        edge["targetLocator"] == "https://github.com/acme/widgets/pull/7"
        for edge in analysis_payload["leadEdges"]
    )
    assert any(source["locator"] == "jira:ABC-2" and source["materialized"] for source in analysis_payload["leadSources"])
    assert any(
        source["locator"] == "confluence:12345" and not source["materialized"]
        for source in analysis_payload["leadSources"]
    )
    assert "-.->|mentions|" in analysis_mermaid or "-.->|mentions x" in analysis_mermaid
    assert "not yet materialized" in analysis_mermaid
    assert any(
        node["id"] == "source:confluence:12345"
        and node["kind"] == "source"
        and node["materialized"] is False
        and node["discovered"] is True
        for node in semantic_payload["nodes"]
    )


def test_session_leads_can_focus_one_artifact_by_artifact_locator(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    source = content.materialize_bytes(
        (
            "# ABC-1\n\n"
            "Depends on ABC-2.\n"
            "PR: https://github.com/acme/widgets/pull/7\n"
        ).encode("utf-8"),
        dirs=dirs,
        preferred_name="ABC-1.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-1",
            "canonical_locator": "jira:ABC-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"# ABC-2\n\nDone.\n",
        dirs=dirs,
        preferred_name="ABC-2.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-2",
            "canonical_locator": "jira:ABC-2",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )
    artifact = content.artifact_locator("ABC-1.md", source.digest)

    monkeypatch.chdir(local_root)

    assert session.main(
        ["leads", "--session", str(local_root), artifact, "--output", "json"]
    ) == 0
    payload = json.loads(capsys.readouterr().out)

    assert payload["artifactCount"] == 1
    assert payload["leadCount"] == 2
    assert payload["artifacts"][0]["artifactLocator"] == artifact
    assert any(
        lead["targetLocator"] == "jira:ABC-2" and lead["materialized"] is True
        for lead in payload["artifacts"][0]["leads"]
    )
    assert any(
        lead["targetLocator"] == "https://github.com/acme/widgets/pull/7"
        and lead["materialized"] is False
        for lead in payload["artifacts"][0]["leads"]
    )


def test_session_leads_orders_best_first_without_quality_thresholds(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"# ABC-1\n\nDepends on ABC-2.\n",
        dirs=dirs,
        preferred_name="ABC-1.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-1",
            "canonical_locator": "jira:ABC-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"# Search Results\n\nhttps://kubernetes.io/docs/reference/networking/virtual-ips/\n",
        dirs=dirs,
        preferred_name="search.md",
        metadata={
            "tool": "gotta",
            "plugin": "slack",
            "locator": "search ABC",
            "canonical_locator": "slack:search ABC",
            "subcommand": "search",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )

    monkeypatch.chdir(local_root)

    assert session.main(["leads", "--session", str(local_root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)

    assert payload["leadCount"] == 2
    assert payload["shownCount"] == 2
    assert payload["materializedLeadCount"] == 0
    assert payload["unmaterializedLeadCount"] == 2
    assert [lead["locator"] for lead in payload["leadSources"]] == [
        "jira:ABC-2",
        "https://kubernetes.io/docs/reference/networking/virtual-ips/",
    ]
    assert payload["leadSources"][0]["firstParty"] is True
    assert payload["leadSources"][0]["materialized"] is False
    assert payload["leadSources"][1]["firstParty"] is False
    assert payload["leadSources"][1]["materialized"] is False
    assert payload["leadSources"][1]["searchLikeSourceCount"] == 1
    search_artifact = next(
        artifact
        for artifact in payload["artifacts"]
        if artifact["sourceLocator"] == "slack:search ABC"
    )
    direct_artifact = next(
        artifact
        for artifact in payload["artifacts"]
        if artifact["sourceLocator"] == "jira:ABC-1"
    )
    assert payload["leadSources"][1]["bestSearchRank"] == 1
    assert payload["leadSources"][1]["searchOrigins"] == [
        {
            "artifactLocator": search_artifact["artifactLocator"],
            "provider": "slack",
            "rank": 1,
            "sourceLocator": "slack:search ABC",
            "subcommand": "search",
        }
    ]
    assert search_artifact["leads"][0]["sourceRank"] == 1
    assert direct_artifact["leads"][0]["sourceRank"] == 0
    assert "minQuality" not in payload
    assert "suppressedLeadCount" not in payload


def test_session_leads_shows_low_signal_only_case_without_hiding_it(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"See also https://kubernetes.io/docs/reference/networking/virtual-ips/\n",
        dirs=dirs,
        preferred_name="search.md",
        metadata={
            "tool": "gotta",
            "plugin": "slack",
            "locator": "search ABC",
            "canonical_locator": "slack:search ABC",
            "subcommand": "search",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )

    monkeypatch.chdir(local_root)

    assert session.main(["leads", "--session", str(local_root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)

    assert payload["leadCount"] == 1
    assert payload["shownCount"] == 1
    assert payload["leadSources"][0]["locator"] == "https://kubernetes.io/docs/reference/networking/virtual-ips/"
    assert payload["leadSources"][0]["firstParty"] is False
    assert payload["leadSources"][0]["materialized"] is False
    assert payload["leadSources"][0]["bestSearchRank"] == 1
    assert payload["leadSources"][0]["searchOrigins"] == [
        {
            "artifactLocator": payload["artifacts"][0]["artifactLocator"],
            "provider": "slack",
            "rank": 1,
            "sourceLocator": "slack:search ABC",
            "subcommand": "search",
        }
    ]
    assert payload["nextStep"] == ""


def test_session_leads_preserve_search_result_order_within_search_artifacts(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        (
            "1. https://kubernetes.io/docs/reference/networking/virtual-ips/\n"
            "2. https://docs.python.org/3/library/pathlib.html\n"
        ).encode("utf-8"),
        dirs=dirs,
        preferred_name="search.md",
        metadata={
            "tool": "gotta",
            "plugin": "confluence",
            "locator": "search networking",
            "canonical_locator": "confluence:search networking",
            "subcommand": "search",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )

    monkeypatch.chdir(local_root)

    assert session.main(["leads", "--session", str(local_root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)

    assert [lead["locator"] for lead in payload["leadSources"]] == [
        "https://kubernetes.io/docs/reference/networking/virtual-ips/",
        "https://docs.python.org/3/library/pathlib.html",
    ]
    assert payload["leadSources"][0]["bestSearchRank"] == 1
    assert payload["leadSources"][1]["bestSearchRank"] == 2
    assert payload["leadSources"][0]["searchOrigins"][0]["provider"] == "confluence"
    assert payload["leadSources"][0]["searchOrigins"][0]["subcommand"] == "search"


def test_session_leads_falls_back_to_same_provider_search_for_prose_heavy_confluence(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        (
            "# Network configuration for debugging devices remotely\n\n"
            "- URL: https://example.atlassian.net/wiki/pages/viewpage.action?pageId=4456054785\n\n"
            "The RMP client or the MeshLink will have a tunnel over cellular connectivity.\n"
        ).encode("utf-8"),
        dirs=dirs,
        preferred_name="4456054785.md",
        metadata={
            "tool": "gotta",
            "plugin": "confluence",
            "locator": "get 4456054785",
            "canonical_locator": "confluence:4456054785",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )

    monkeypatch.chdir(local_root)

    assert session.main(["leads", "--session", str(local_root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)

    locators = {lead["locator"] for lead in payload["leadSources"]}
    assert "confluence:search RMP MeshLink" in locators
    confluence_lead = next(
        lead
        for lead in payload["leadSources"]
        if lead["locator"] == "confluence:search RMP MeshLink"
    )
    assert confluence_lead["firstParty"] is True
    assert confluence_lead["materialized"] is False
    assert confluence_lead["searchSeed"] is True
    assert confluence_lead["relationKinds"] == ["suggests_search"]


def test_session_leads_falls_back_to_workspace_scoped_slack_search_for_semantic_threads(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        (
            "### Slack Thread: RMP retrospective\n\n"
            "- _Source_: https://demo.slack.com/archives/C12345678/p1773085070240949\n\n"
            "- _2026-03-13 14:05:26.726_ **operator**: "
            "@teammate is always elevated as the one who reined in RMP "
            "via the edge services change lol\n"
        ).encode("utf-8"),
        dirs=dirs,
        preferred_name="p1773085070240949.md",
        metadata={
            "tool": "gotta",
            "plugin": "slack",
            "locator": "get C12345678:1773085070.240949",
            "canonical_locator": "slack:thread:C12345678:1773085070240949",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )

    monkeypatch.chdir(local_root)

    assert session.main(["leads", "--session", str(local_root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)

    locators = {lead["locator"] for lead in payload["leadSources"]}
    assert "slack:search --workspace demo RMP services" in locators
    slack_lead = next(
        lead
        for lead in payload["leadSources"]
        if lead["locator"] == "slack:search --workspace demo RMP services"
    )
    assert slack_lead["firstParty"] is True
    assert slack_lead["materialized"] is False
    assert slack_lead["searchSeed"] is True
    assert slack_lead["relationKinds"] == ["suggests_search"]


def test_extract_explicit_leads_strips_wrapped_url_noise() -> None:
    mentions = leads.extract_explicit_leads(
        "PR: <https://github.com/acme/widgets/pull/7|PR 7>\n"
        "Issue: https://github.com/acme/widgets/issues/9)*\n"
    )

    locators = {mention.canonical_locator for mention in mentions}
    assert "https://github.com/acme/widgets/pull/7" in locators
    assert "https://github.com/acme/widgets/issues/9" in locators
    assert all("|" not in locator for locator in locators)


def test_extract_explicit_leads_drops_partial_atlassian_browse_urls() -> None:
    mentions = leads.extract_explicit_leads(
        "Truncated: <https://example.atlassian.net/browse/AB|AB>\n"
        "Exact: <https://example.atlassian.net/browse/ABC-12|ABC-12>\n"
    )

    locators = {mention.canonical_locator for mention in mentions}
    assert "jira:ABC-12" in locators
    assert "https://example.atlassian.net/browse/AB" not in locators


def test_extract_explicit_leads_drops_malformed_atlassian_wiki_embed_urls() -> None:
    mentions = leads.extract_explicit_leads(
        "2100446110108744560547851Untitled Diagram-1773161260589.drawio22"
        "https://example.atlassian.net/wikiUntitled Diagram-1773161260589.drawio01170.5521\n"
    )

    locators = {mention.canonical_locator for mention in mentions}
    assert "https://example.atlassian.net/wikiUntitled" not in locators


def test_extract_explicit_leads_drops_obvious_placeholder_and_asset_urls() -> None:
    mentions = leads.extract_explicit_leads(
        "Badge: https://img.shields.io/badge/build-passing-brightgreen.svg\n"
        "Local example: http://127.0.0.1:7400\n"
        "Placeholder: http://www.example.com:8080\n"
        "Real doc: https://kubernetes.io/docs/reference/networking/virtual-ips/\n"
    )

    locators = {mention.canonical_locator for mention in mentions}
    assert "https://kubernetes.io/docs/reference/networking/virtual-ips/" in locators
    assert "https://img.shields.io/badge/build-passing-brightgreen.svg" not in locators
    assert "http://127.0.0.1:7400" not in locators
    assert "http://www.example.com:8080" not in locators


def test_extract_explicit_leads_prefers_root_thread_from_slack_reply_permalink() -> None:
    mentions = leads.extract_explicit_leads(
        "Reply: https://demo.slack.com/archives/C12345678/p1773081279142849?thread_ts=1773075428.384009\n"
    )

    locators = {mention.canonical_locator for mention in mentions}
    assert "slack:thread:C12345678:1773075428384009" in locators
    assert "slack:thread:C12345678:1773081279142849" not in locators


def test_extract_explicit_leads_drops_ellipsized_host_only_urls() -> None:
    mentions = leads.extract_explicit_leads(
        "#### Hi there <https://kubernetes.i...\n"
        "Real: https://kubernetes.io/docs/concepts/services-networking/service/\n"
    )

    locators = {mention.canonical_locator for mention in mentions}
    assert "https://kubernetes.i" not in locators
    assert "https://kubernetes.io/docs/concepts/services-networking/service/" in locators


def test_materialize_bytes_eagerly_writes_lead_cache(tmp_path: Path, monkeypatch) -> None:
    dirs = initialize_session(tmp_path / "local")
    result = content.materialize_bytes(
        b"Depends on ABC-2.\nDesign doc: confluence:12345\n",
        dirs=dirs,
        preferred_name="ABC-1.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-1",
            "canonical_locator": "jira:ABC-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    cache_path = result.content_dir / "leads.json"

    assert cache_path.exists()
    payload = json.loads(cache_path.read_text(encoding="utf-8"))
    assert payload["version"] == leads.LEADS_CACHE_VERSION
    assert {entry["canonical_locator"] for entry in payload["entries"]} == {
        "jira:ABC-2",
        "confluence:12345",
    }

    snapshot = content.scan_content_store(dirs.content_dir)[0]

    def fail_extract(_text: str) -> list[leads.LeadMention]:
        raise AssertionError("lead extraction should reuse the eager cache")

    monkeypatch.setattr(leads, "extract_explicit_leads", fail_extract)
    mentions = leads.lead_mentions_for_snapshot(snapshot)

    assert {mention.canonical_locator for mention in mentions} == {
        "jira:ABC-2",
        "confluence:12345",
    }


def test_lead_mentions_for_snapshot_rebuilds_stale_cache(tmp_path: Path) -> None:
    dirs = initialize_session(tmp_path / "local")
    result = content.materialize_bytes(
        b"Depends on ABC-2.\n",
        dirs=dirs,
        preferred_name="ABC-1.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "get ABC-1",
            "canonical_locator": "jira:ABC-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    cache_path = result.content_dir / "leads.json"
    cache_path.write_text(
        json.dumps({"version": 0, "entries": []}) + "\n",
        encoding="utf-8",
    )

    snapshot = content.scan_content_store(dirs.content_dir)[0]
    mentions = leads.lead_mentions_for_snapshot(snapshot)

    assert [mention.canonical_locator for mention in mentions] == ["jira:ABC-2"]
    payload = json.loads(cache_path.read_text(encoding="utf-8"))
    assert payload["version"] == leads.LEADS_CACHE_VERSION
    assert payload["leadCount"] == 1


def test_session_manifest_has_native_summary_surface(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"hello world",
        dirs=dirs,
        preferred_name="demo.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": "demo",
            "canonical_locator": "demo",
            "actor": "claude",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    assert session.main(
        ["manifest", "--session", str(local_root), "--actor", "claude", "--output", "json"]
    ) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entryCount"] == 1
    assert payload["entries"][0]["actor"] == "claude"
    assert payload["entries"][0]["canonical_locator"] == "demo"
    assert payload["entries"][0]["follow_command"] == "gotta read 'demo'"
    assert payload["entries"][0]["content_locator"].startswith("content:")
    assert payload["entries"][0]["artifact_locator"].startswith("artifact:demo.md@")
    assert payload["entries"][0]["content_follow_command"].startswith("gotta read 'content:")
    assert payload["entries"][0]["artifact_follow_command"].startswith("gotta read 'artifact:demo.md@")


def test_session_manifest_accepts_stdout_flag_for_uniformity(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"hello world",
        dirs=dirs,
        preferred_name="demo.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": "demo",
            "canonical_locator": "demo",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    assert session.main(["manifest", "--session", str(local_root), "--output", "json", "--stdout"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entryCount"] == 1


def test_session_manifest_falls_back_to_content_locator_when_canonical_missing(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    result = content.materialize_bytes(
        b"hello world",
        dirs=dirs,
        preferred_name="demo.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": "",
            "canonical_locator": "",
            "actor": "claude",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    assert session.main(
        ["manifest", "--session", str(local_root), "--actor", "claude", "--output", "json"]
    ) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entries"][0]["follow_command"] == f"gotta read 'content:{result.digest}'"
    assert payload["entries"][0]["content_locator"] == f"content:{result.digest}"
    assert payload["entries"][0]["artifact_locator"] == content.artifact_locator("demo.md", result.digest)


def test_session_timeline_has_native_continuity_surface(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"first",
        dirs=dirs,
        preferred_name="first.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"second",
        dirs=dirs,
        preferred_name="second.md",
        metadata={
            "tool": "gotta",
            "plugin": "confluence",
            "locator": "3904339970",
            "canonical_locator": "confluence:3904339970",
            "actor": "claude",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )
    assert session.main(["timeline", "--session", str(local_root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["mode"] == "acquired"
    assert payload["eventCount"] == 2
    assert payload["events"][0]["mode"] == "acquired"
    assert payload["events"][0]["follow_command"] == "gotta read 'jira:PROJ-1'"
    assert payload["events"][0]["locator"] == "jira:PROJ-1"
    assert payload["events"][0]["content_locator"].startswith("content:")
    assert payload["events"][0]["artifact_locator"].startswith("artifact:first.md@")
    assert payload["events"][1]["actor"] == "claude"


def test_session_timeline_accepts_stdout_flag_for_uniformity(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"first",
        dirs=dirs,
        preferred_name="first.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    assert session.main(["timeline", "--session", str(local_root), "--output", "json", "--stdout"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["eventCount"] == 1


def test_session_timeline_acquired_includes_native_local_activity(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="PROJ-1.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.append_activity_event(
        local_root,
        {
            "timestamp": "2026-03-11T00:00:01Z",
            "plugin": "logs",
            "surface": "logs",
            "action": "append",
            "locator": "LOGS.md",
            "preferred_name": "LOGS.md",
            "follow_command": "gotta read 'LOGS.md'",
            "detail": "appended 1 logs entry",
            "time_field": "session_recorded_at",
        },
    )
    assert session.main(["timeline", "--session", str(local_root), "--output", "json"]) == 0

    payload = json.loads(capsys.readouterr().out)
    assert payload["activityPath"].endswith("state/activity.jsonl")
    assert payload["eventCount"] == 2
    assert payload["events"][0]["locator"] == "jira:PROJ-1"
    assert payload["events"][1]["locator"] == "LOGS.md"
    assert payload["events"][1]["event_kind"] == "local"
    assert payload["events"][1]["fetched_at"] == "2026-03-11T00:00:01Z"


def test_session_timeline_best_effort_mode_prefers_created_and_surfaces_gaps(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="jira.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
            "created": "2026-03-09T09:00:00Z",
            "updated": "2026-03-10T12:00:00Z",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"readme body",
        dirs=dirs,
        preferred_name="README.md",
        metadata={
            "tool": "gotta",
            "plugin": "github",
            "locator": "https://github.com/example/repo",
            "canonical_locator": "https://github.com/example/repo",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )
    assert (
        session.main(
            ["timeline", "--session", str(local_root), "--mode", "best-effort", "--output", "json"]
        )
        == 0
    )
    payload = json.loads(capsys.readouterr().out)
    assert payload["mode"] == "best-effort"
    assert payload["coverageGapCount"] == 1
    assert payload["eventCount"] == 1
    assert payload["events"][0]["source_time_field"] == "source_created_at"
    assert payload["events"][0]["source_time"].startswith("2026-03-09T09:00:00")
    assert payload["events"][0]["source_created_at"].startswith("2026-03-09T09:00:00")
    assert payload["events"][0]["source_updated_at"].startswith("2026-03-10T12:00:00")


def test_session_timeline_best_effort_includes_local_activity_with_explicit_provenance(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="jira.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
            "created": "2026-03-09T09:00:00Z",
            "updated": "2026-03-10T12:00:00Z",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.append_activity_event(
        local_root,
        {
            "timestamp": "2026-03-11T00:00:01Z",
            "plugin": "todo",
            "surface": "todo",
            "action": "append",
            "locator": "TODO.md",
            "preferred_name": "TODO.md",
            "follow_command": "gotta read 'TODO.md'",
            "detail": "appended 1 TODO item",
            "time_field": "session_recorded_at",
        },
    )
    assert (
        session.main(
            ["timeline", "--session", str(local_root), "--mode", "best-effort", "--output", "json"]
        )
        == 0
    )

    payload = json.loads(capsys.readouterr().out)
    assert payload["coverageGapCount"] == 0
    assert payload["eventCount"] == 2
    assert payload["events"][0]["locator"] == "jira:PROJ-1"
    assert payload["events"][0]["source_time_field"] == "source_created_at"
    assert payload["events"][1]["locator"] == "TODO.md"
    assert payload["events"][1]["source_time_field"] == "session_recorded_at"
    assert payload["events"][1]["event_kind"] == "local"


def test_session_timeline_created_and_updated_modes_split_cleanly(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="jira.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
            "created": "2026-03-09T09:00:00Z",
            "updated": "2026-03-10T12:00:00Z",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    assert session.main(["timeline", "--session", str(local_root), "--mode", "created", "--output", "json"]) == 0
    created_payload = json.loads(capsys.readouterr().out)
    assert created_payload["mode"] == "created"
    assert created_payload["coverageGapCount"] == 0
    assert created_payload["events"][0]["source_time_field"] == "source_created_at"
    assert created_payload["events"][0]["source_time"].startswith("2026-03-09T09:00:00")

    assert session.main(["timeline", "--session", str(local_root), "--mode", "updated", "--output", "json"]) == 0
    updated_payload = json.loads(capsys.readouterr().out)
    assert updated_payload["mode"] == "updated"
    assert updated_payload["coverageGapCount"] == 0
    assert updated_payload["events"][0]["source_time_field"] == "source_updated_at"
    assert updated_payload["events"][0]["source_time"].startswith("2026-03-10T12:00:00")

def test_session_timeline_source_mode_ignores_local_reads_and_schema_probes(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="jira.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
            "created": "2026-03-09T09:00:00Z",
            "updated": "2026-03-10T12:00:00Z",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"goal body",
        dirs=dirs,
        preferred_name="GOAL.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": "GOAL.md",
            "canonical_locator": "GOAL.md",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )
    content.materialize_bytes(
        b"schema body",
        dirs=dirs,
        preferred_name="workspace.tsv",
        metadata={
            "tool": "gotta",
            "plugin": "slack",
            "locator": "slack:sql --workspace demo --output tsv PRAGMA table_info(MESSAGE)",
            "canonical_locator": "slack:sql --workspace demo --output tsv PRAGMA table_info(MESSAGE)",
        },
        timestamp="2026-03-11T00:00:02.000001Z",
    )
    assert (
        session.main(
            ["timeline", "--session", str(local_root), "--mode", "best-effort", "--output", "json"]
        )
        == 0
    )
    payload = json.loads(capsys.readouterr().out)
    assert payload["coverageGapCount"] == 0
    assert payload["eventCount"] == 1
    assert payload["events"][0]["locator"] == "jira:PROJ-1"


def test_session_timeline_best_effort_ignores_local_artifact_rereads(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    source = content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="jira.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
            "created": "2026-03-09T09:00:00Z",
            "updated": "2026-03-10T12:00:00Z",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"# Jira\n\nsnippet\n",
        dirs=dirs,
        preferred_name="jira-head.md",
        metadata={
            "tool": "gotta",
            "plugin": "read",
            "locator": f"artifact:jira.md@{source.digest[:12]} --head 20",
            "canonical_locator": f"artifact:jira.md@{source.digest[:12]} --head 20",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )
    assert (
        session.main(
            ["timeline", "--session", str(local_root), "--mode", "best-effort", "--output", "json"]
        )
        == 0
    )
    payload = json.loads(capsys.readouterr().out)
    assert payload["coverageGapCount"] == 0
    assert payload["eventCount"] == 1
    assert payload["events"][0]["locator"] == "jira:PROJ-1"


def test_session_timeline_best_effort_ignores_aggregate_search_artifact_dates(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    content.materialize_bytes(
        b"# Search Results\n\n- hit\n",
        dirs=dirs,
        preferred_name="gdocs-search-abc.md",
        metadata={
            "tool": "gotta",
            "plugin": "gdocs",
            "subcommand": "search",
            "locator": "search --limit 8 ABC",
            "canonical_locator": "gdocs:search --limit 8 ABC",
            "source_created_at": "2025-10-03T18:43:19.172Z",
            "source_updated_at": "2026-02-25T19:24:20.112Z",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )

    assert (
        session.main(
            ["timeline", "--session", str(local_root), "--mode", "best-effort", "--output", "json"]
        )
        == 0
    )
    best_effort_payload = json.loads(capsys.readouterr().out)
    assert best_effort_payload["coverageGapCount"] == 0
    assert best_effort_payload["eventCount"] == 0

    assert (
        session.main(
            ["timeline", "--session", str(local_root), "--mode", "created", "--output", "json"]
        )
        == 0
    )
    created_payload = json.loads(capsys.readouterr().out)
    assert created_payload["coverageGapCount"] == 0
    assert created_payload["eventCount"] == 0


def test_session_manifest_plugin_filter_sees_provider_attributed_read_artifacts(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    local_root = tmp_path / "local"
    dirs = initialize_session(local_root)
    dispatch._materialize_invocation(
        "read",
        ["jira:PROJ-1"],
        content.CommonOptions(),
        b"# PROJ-1\n\n- Created: 2026-03-10T12:00:00Z\n",
        dirs=dirs,
    )
    assert session.main(["manifest", "--session", str(local_root), "--plugin", "jira", "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entryCount"] == 1
    assert payload["entries"][0]["plugin"] == "jira"
    assert payload["entries"][0]["canonical_locator"] == "jira:PROJ-1"


def test_session_analyze_does_not_treat_same_name_cross_provider_as_revision(
    tmp_path: Path
) -> None:
    dirs = initialize_session(tmp_path / "local")
    content.materialize_bytes(
        b"jira body",
        dirs=dirs,
        preferred_name="ABC.md",
        metadata={
            "tool": "gotta",
            "plugin": "jira",
            "locator": "PROJ-1",
            "canonical_locator": "jira:PROJ-1",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"confluence body",
        dirs=dirs,
        preferred_name="ABC.md",
        metadata={
            "tool": "gotta",
            "plugin": "confluence",
            "locator": "3904339970",
            "canonical_locator": "confluence:3904339970",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )

    payload = session._analysis_payload(dirs)

    assert payload["revisionEdgeCount"] == 0
    provider_sets = {tuple(item["providers"]) for item in payload["content"]}
    assert ("confluence",) in provider_sets
    assert ("jira",) in provider_sets


def test_session_analyze_marks_same_name_collisions_with_resource_hints(
    tmp_path: Path
) -> None:
    dirs = initialize_session(tmp_path / "local")
    content.materialize_bytes(
        b"thread one",
        dirs=dirs,
        preferred_name="shared-thread.md",
        metadata={
            "tool": "gotta",
            "plugin": "slack",
            "locator": "get https://example.slack.com/archives/C1/p111",
            "canonical_locator": "slack:thread:C1:111",
        },
        timestamp="2026-03-11T00:00:00.000001Z",
    )
    content.materialize_bytes(
        b"thread two",
        dirs=dirs,
        preferred_name="shared-thread.md",
        metadata={
            "tool": "gotta",
            "plugin": "slack",
            "locator": "get https://example.slack.com/archives/C1/p222",
            "canonical_locator": "slack:thread:C1:222",
        },
        timestamp="2026-03-11T00:00:01.000001Z",
    )

    payload = session._analysis_payload(dirs)

    assert payload["nameCollisionCount"] == 1
    assert payload["nameCollisions"] == ["shared-thread.md"]
    assert all(item["nameCollision"] for item in payload["content"])
    assert any(
        any("slack-thread:" in hint for hint in item["resourceHints"])
        for item in payload["content"]
    )

    mermaid = session._render_analysis_mermaid(payload)
    assert "slack-thread:C1:111" in mermaid or "slack-thread:C1:222" in mermaid



"""Shared user-config and state-path helpers."""

from __future__ import annotations

from collections.abc import Mapping
import json
import os
from pathlib import Path
from typing import Any
from platformdirs import PlatformDirs

from gotta.compat import tomllib


APP_NAME = "gotta"
CONFIG_FILE_NAME = "gotta.toml"


def _dirs() -> PlatformDirs:
    return PlatformDirs(appname=APP_NAME, appauthor=False, ensure_exists=False)


def user_config_dir() -> Path:
    override = os.environ.get("GOTTA_CONFIG_DIR", "").strip()
    if override:
        return Path(override).expanduser()
    return Path(_dirs().user_config_path)


def user_data_dir() -> Path:
    override = os.environ.get("GOTTA_DATA_DIR", "").strip()
    if override:
        return Path(override).expanduser()
    return Path(_dirs().user_data_path)


def user_state_dir() -> Path:
    override = os.environ.get("GOTTA_STATE_DIR", "").strip()
    if override:
        return Path(override).expanduser()
    return Path(_dirs().user_state_path)


def default_config_file() -> Path:
    return user_config_dir() / CONFIG_FILE_NAME


def primary_config_file() -> Path:
    override = os.environ.get("GOTTA_CONFIG_FILE", "").strip()
    if override:
        return Path(override).expanduser()
    return default_config_file()


def config_file_candidates() -> list[Path]:
    return [primary_config_file()]


def display_path(path: Path) -> str:
    try:
        return "~/" + str(path.expanduser().relative_to(Path.home()))
    except ValueError:
        return str(path)


def provider_env_reference(provider: str) -> str:
    return (
        f"[providers.{provider}.env] in GOTTA_CONFIG_FILE "
        f"(default {display_path(default_config_file())})"
    )


def env_or_config(config_env: Mapping[str, str], *names: str, default: str = "") -> str:
    for name in names:
        value = os.environ.get(name, "").strip()
        if value:
            return value
    for name in names:
        value = str(config_env.get(name, "")).strip()
        if value:
            return value
    return default


def extract_provider_env(config: Mapping[str, Any], provider: str) -> dict[str, str]:
    providers = config.get("providers")
    if isinstance(providers, Mapping):
        provider_table = providers.get(provider)
        if isinstance(provider_table, Mapping):
            env_table = provider_table.get("env")
            if isinstance(env_table, Mapping):
                result = {
                    key: value
                    for key, value in env_table.items()
                    if isinstance(key, str) and isinstance(value, str)
                }
                if result:
                    return result
    return {}


def load_config() -> dict[str, Any]:
    path = primary_config_file()
    if not path.exists():
        return {}
    try:
        data = tomllib.loads(path.read_text(encoding="utf-8"))
    except tomllib.TOMLDecodeError:
        return {}
    return data if isinstance(data, dict) else {}


def _toml_scalar(value: Any) -> str:
    if isinstance(value, str):
        return json.dumps(value)
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)) and not isinstance(value, bool):
        return str(value)
    raise TypeError(f"unsupported config value: {value!r}")


def _render_table(prefix: list[str], table: Mapping[str, Any], lines: list[str]) -> None:
    scalars = [
        (key, value)
        for key, value in table.items()
        if not isinstance(value, Mapping)
    ]
    subtables = [
        (key, value)
        for key, value in table.items()
        if isinstance(value, Mapping)
    ]
    if prefix:
        lines.append(f"[{'.'.join(prefix)}]")
        for key, value in scalars:
            lines.append(f"{key} = {_toml_scalar(value)}")
        lines.append("")
    for key, value in subtables:
        _render_table([*prefix, key], value, lines)


def write_config(config: Mapping[str, Any]) -> Path:
    path = primary_config_file()
    path.parent.mkdir(parents=True, exist_ok=True)
    lines: list[str] = []
    _render_table([], config, lines)
    text = "\n".join(lines).rstrip() + "\n" if lines else ""
    path.write_text(text, encoding="utf-8")
    return path


def set_provider_env_values(provider: str, updates: Mapping[str, str]) -> Path:
    config = load_config()
    providers = config.setdefault("providers", {})
    if not isinstance(providers, dict):
        providers = {}
        config["providers"] = providers
    provider_table = providers.setdefault(provider, {})
    if not isinstance(provider_table, dict):
        provider_table = {}
        providers[provider] = provider_table
    env_table = provider_table.setdefault("env", {})
    if not isinstance(env_table, dict):
        env_table = {}
        provider_table["env"] = env_table
    for key, value in updates.items():
        env_table[key] = value
    return write_config(config)



"""Canonical structured session log state and Markdown projection."""

from __future__ import annotations

from pathlib import Path

from gotta.compat import UTC, datetime
from gotta.content import load_state_env_at_root, write_text_atomic
from gotta.projection import append_chunk, append_jsonl, read_jsonl_records


LOGS_LOG_NAME = "logs.jsonl"


def logs_state_path(work_dir: Path) -> Path:
    return work_dir / "state" / LOGS_LOG_NAME


def logs_surface_path(work_dir: Path) -> Path:
    return work_dir / "LOGS.md"

def log_records(work_dir: Path) -> list[dict[str, object]]:
    return read_jsonl_records(logs_state_path(work_dir))


def logs_payload(work_dir: Path, *, limit: int = 0) -> dict[str, object]:
    records = sorted(log_records(work_dir), key=lambda item: str(item.get("timestamp") or ""))
    entries = records[-limit:] if limit > 0 else records
    return {
        "logs": str(logs_surface_path(work_dir)),
        "logs_log": str(logs_state_path(work_dir)),
        "entry_count": len(records),
        "entries": entries,
    }


def _repo_display_name(work_dir: Path) -> str:
    state = load_state_env_at_root(work_dir)
    repo_path = str(state.get("GOTTA_WORK_REPO") or "").strip()
    if not repo_path:
        return "Session"
    return Path(repo_path).name.capitalize() or "Session"


def render_logs_markdown(work_dir: Path, records: list[dict[str, object]]) -> str:
    lines = [
        "# Logs",
        "",
        f"> Generated automatically from `state/{LOGS_LOG_NAME}`.",
        "> This is a human-readable projection; the structured `logs` log is canonical.",
        "> Prefer `gotta logs ...` for routine mutation and inspection.",
        "",
        f"This is the durable execution log for the {_repo_display_name(work_dir)} session.",
        "",
        "## Entries",
        "",
    ]
    ordered = sorted(records, key=lambda item: str(item.get("timestamp") or ""))
    for record in ordered:
        timestamp = str(record.get("timestamp") or "unknown-time")
        message = str(record.get("message") or "").strip() or "unspecified log entry"
        message_lines = message.splitlines() or ["unspecified log entry"]
        lines.append(f"- `{timestamp}` {message_lines[0]}")
        for continuation in message_lines[1:]:
            lines.append(f"  {continuation}")
    return "\n".join(lines) + "\n"


def render_log_record_markdown(record: dict[str, object]) -> str:
    timestamp = str(record.get("timestamp") or "unknown-time")
    message = str(record.get("message") or "").strip() or "unspecified log entry"
    message_lines = message.splitlines() or ["unspecified log entry"]
    lines = [f"- `{timestamp}` {message_lines[0]}"]
    for continuation in message_lines[1:]:
        lines.append(f"  {continuation}")
    return "\n".join(lines) + "\n"


def sync_logs_projection(work_dir: Path) -> None:
    write_text_atomic(
        logs_surface_path(work_dir),
        render_logs_markdown(work_dir, log_records(work_dir)),
    )


def append_log_record(
    work_dir: Path,
    *,
    message: str,
    actor: str = "primary",
    timestamp: str | None = None,
) -> dict[str, object]:
    payload: dict[str, object] = {
        "timestamp": timestamp or datetime.now(tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
        "message": message,
        "actor": actor,
        "channel": "logs",
    }
    log_path = logs_state_path(work_dir)
    surface_path = logs_surface_path(work_dir)
    append_jsonl(log_path, payload)
    if surface_path.exists():
        append_chunk(surface_path, render_log_record_markdown(payload))
    else:
        sync_logs_projection(work_dir)
    return payload


def recent_log_lines(work_dir: Path, *, limit: int) -> list[str]:
    records = sorted(log_records(work_dir), key=lambda item: str(item.get("timestamp") or ""))
    entries: list[str] = []
    for record in records[-limit:]:
        timestamp = str(record.get("timestamp") or "unknown-time")
        message = str(record.get("message") or "").strip() or "unspecified log entry"
        first_line = message.splitlines()[0] if message.splitlines() else message
        entries.append(f"- `{timestamp}` {first_line}")
    return entries



"""Runtime actor context for the `gotta ...` operator frame."""

from __future__ import annotations

from dataclasses import dataclass
import os


ACTOR_SPEAKER_ENV = "GOTTA_ACTOR_SPEAKER"
ACTOR_CALLEE_ENV = "GOTTA_ACTOR_CALLEE"

PRIMARY_ACTOR = "primary"
USER_ACTOR = "user"


@dataclass(frozen=True, slots=True)
class ActorContext:
    speaker: str
    callee: str


def resolve_actor_context() -> ActorContext:
    speaker = str(os.environ.get(ACTOR_SPEAKER_ENV) or "").strip() or PRIMARY_ACTOR
    callee = str(os.environ.get(ACTOR_CALLEE_ENV) or "").strip() or USER_ACTOR
    return ActorContext(speaker=speaker, callee=callee)


def seed_primary_actor_context() -> None:
    os.environ.setdefault(ACTOR_SPEAKER_ENV, PRIMARY_ACTOR)
    os.environ.setdefault(ACTOR_CALLEE_ENV, USER_ACTOR)



"""Built-in plugin contract, discovery, and registrations."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from functools import cache
import importlib
from importlib.metadata import entry_points
import inspect
import sys
from typing import Any


Runner = Callable[[list[str]], int]
RouteTarget = Callable[[str], list[str] | None]
ShouldMaterialize = Callable[[list[str]], bool]
InvocationLocator = Callable[[list[str]], str]
CanonicalLocator = Callable[[list[str]], str]
PreferredName = Callable[[list[str], Any], str]
ContentType = Callable[[list[str], str], str]

DEFAULT_PLUGIN_GROUP = "gotta.plugins"
ASK_PLUGIN_GROUP = "gotta.ask"


@dataclass(frozen=True, slots=True)
class PluginSpec:
    name: str
    description: str
    runner: Runner
    route_target: RouteTarget | None = None
    route_priority: int = 100
    should_materialize: ShouldMaterialize | None = None
    invocation_locator: InvocationLocator | None = None
    canonical_locator: CanonicalLocator | None = None
    preferred_name: PreferredName | None = None
    content_type: ContentType | None = None


CORE_PLUGIN_DISTS = {"gotta"}


def _is_zero_arg_factory(value: Callable[..., object]) -> bool:
    signature = inspect.signature(value)
    required = [
        parameter
        for parameter in signature.parameters.values()
        if parameter.default is inspect.Signature.empty
        and parameter.kind
        in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        )
    ]
    return not required


def _coerce_plugin(name: str, loaded: object) -> PluginSpec:
    if isinstance(loaded, PluginSpec):
        return loaded
    if callable(loaded) and _is_zero_arg_factory(loaded):
        produced = loaded()
        if isinstance(produced, PluginSpec):
            return produced
    if callable(loaded):
        return PluginSpec(name=name, description="", runner=loaded)
    runner = getattr(loaded, "main", None)
    if callable(runner):
        return PluginSpec(name=name, description="", runner=runner)
    raise RuntimeError(
        f"plugin '{name}' entry point must load a PluginSpec, a factory returning "
        "PluginSpec, a callable runner, or an object with main()"
    )


def _dist_name(entry_point: Any) -> str:
    dist = getattr(entry_point, "dist", None)
    name = getattr(dist, "name", "")
    return str(name or "")


def _shadow_priority(entry_point: Any) -> tuple[int, str, str]:
    dist_name = _dist_name(entry_point)
    is_external = 1 if dist_name not in CORE_PLUGIN_DISTS else 0
    return (is_external, dist_name, str(getattr(entry_point, "value", "")))


def _warn_broken_entry_point(entry_point: Any, group: str, exc: Exception) -> None:
    dist_name = _dist_name(entry_point) or "<unknown-dist>"
    print(
        "warning: ignoring broken plugin entry point "
        f"{group}:{entry_point.name} from {dist_name}: "
        f"{type(exc).__name__}: {exc}",
        file=sys.stderr,
    )


@cache
def discovered_plugins(group: str = DEFAULT_PLUGIN_GROUP) -> dict[str, PluginSpec]:
    discovered: dict[str, PluginSpec] = (
        _builtin_core_plugins() if group == DEFAULT_PLUGIN_GROUP else {}
    )
    for entry_point in sorted(entry_points(group=group), key=_shadow_priority):
        if group == DEFAULT_PLUGIN_GROUP and _dist_name(entry_point) in CORE_PLUGIN_DISTS:
            # Source-defined core plugins are canonical. When running from source,
            # installed core metadata may lag behind and still advertise removed
            # plugins; skip all core entry-point metadata instead of warning on it.
            continue
        try:
            plugin = _coerce_plugin(entry_point.name, entry_point.load())
        except Exception as exc:
            # Stale or broken external entry-point metadata must not break the core CLI.
            _warn_broken_entry_point(entry_point, group, exc)
            continue
        if plugin.name != entry_point.name:
            plugin = PluginSpec(
                name=entry_point.name,
                description=plugin.description,
                runner=plugin.runner,
                route_target=plugin.route_target,
                route_priority=plugin.route_priority,
                should_materialize=plugin.should_materialize,
                invocation_locator=plugin.invocation_locator,
                canonical_locator=plugin.canonical_locator,
                preferred_name=plugin.preferred_name,
                content_type=plugin.content_type,
            )
        discovered[plugin.name] = plugin
    return discovered


def clear_plugin_cache() -> None:
    discovered_plugins.cache_clear()


def get_plugin(name: str, *, group: str = DEFAULT_PLUGIN_GROUP) -> PluginSpec | None:
    return discovered_plugins(group).get(name)


def available_plugins(*, group: str = DEFAULT_PLUGIN_GROUP) -> list[str]:
    return sorted(discovered_plugins(group))


def iter_plugins(*, group: str = DEFAULT_PLUGIN_GROUP) -> list[PluginSpec]:
    return [discovered_plugins(group)[name] for name in available_plugins(group=group)]


def _runner(module_name: str):
    def run(argv: list[str]) -> int:
        module = importlib.import_module(module_name)
        tool_error = getattr(module, "ToolError", None)
        die = getattr(module, "die", None)
        try:
            return int(module.main(argv))
        except BrokenPipeError:
            raise
        except Exception as exc:
            if isinstance(tool_error, type) and isinstance(exc, tool_error) and callable(die):
                return int(die(str(exc)))
            raise

    return run


def _module_attr(module_name: str, attr: str):
    def call(*args):
        module = importlib.import_module(module_name)
        return getattr(module, attr)(*args)

    return call


def confluence_plugin() -> PluginSpec:
    return PluginSpec(
        name="confluence",
        description="read and edit Confluence pages with durable OAuth-backed access",
        runner=_runner("gotta.plugins.confluence"),
        route_target=_module_attr("gotta.plugins.confluence", "route_target"),
        route_priority=20,
        canonical_locator=_module_attr("gotta.plugins.confluence", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.confluence", "preferred_name"),
    )


def ask_plugin() -> PluginSpec:
    return PluginSpec(
        name="ask",
        description="dispatch to installed ask-family extensions",
        runner=_runner("gotta.plugins.ask"),
        should_materialize=_module_attr("gotta.plugins.ask", "should_materialize"),
        invocation_locator=_module_attr("gotta.plugins.ask", "invocation_locator"),
        canonical_locator=_module_attr("gotta.plugins.ask", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.ask", "preferred_name"),
        content_type=_module_attr("gotta.plugins.ask", "content_type"),
    )


def read_plugin() -> PluginSpec:
    return PluginSpec(
        name="read",
        description="acquire one target through the right retrieval surface and materialize the result",
        runner=_runner("gotta.plugins.read"),
        should_materialize=_module_attr("gotta.target", "should_materialize"),
        canonical_locator=_module_attr("gotta.target", "canonical_locator"),
        preferred_name=_module_attr("gotta.target", "preferred_name"),
    )


def oops_plugin() -> PluginSpec:
    return PluginSpec(
        name="oops",
        description="capture and inspect durable session friction",
        runner=_runner("gotta.plugins.oops"),
        should_materialize=lambda argv: False,
    )


def todo_plugin() -> PluginSpec:
    return PluginSpec(
        name="todo",
        description="inspect and mutate the canonical session checklist",
        runner=_runner("gotta.plugins.todo"),
        should_materialize=lambda argv: False,
    )


def want_plugin() -> PluginSpec:
    return PluginSpec(
        name="want",
        description="inspect or rewrite the canonical session intent frame",
        runner=_runner("gotta.plugins.want"),
        should_materialize=lambda argv: False,
    )


def logs_plugin() -> PluginSpec:
    return PluginSpec(
        name="logs",
        description="inspect and mutate the canonical session execution log",
        runner=_runner("gotta.plugins.logs"),
        should_materialize=lambda argv: False,
    )


def goal_plugin() -> PluginSpec:
    return PluginSpec(
        name="goal",
        description="inspect or rewrite the canonical session goal charter",
        runner=_runner("gotta.plugins.goal"),
        should_materialize=lambda argv: False,
    )


def notes_plugin() -> PluginSpec:
    return PluginSpec(
        name="notes",
        description="inspect and mutate canonical peer notes inside the active session",
        runner=_runner("gotta.plugins.notes"),
        should_materialize=lambda argv: False,
    )


def peer_plugin() -> PluginSpec:
    return PluginSpec(
        name="peer",
        description="configure, launch, and disposition linked peer sessions inside the active session",
        runner=_runner("gotta.plugins.peer"),
        should_materialize=lambda argv: False,
    )


def gdocs_plugin() -> PluginSpec:
    return PluginSpec(
        name="gdocs",
        description="read and search Google Docs through the Drive/Docs APIs",
        runner=_runner("gotta.plugins.gdocs"),
        route_target=_module_attr("gotta.plugins.gdocs", "route_target"),
        route_priority=50,
        canonical_locator=_module_attr("gotta.plugins.gdocs", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.gdocs", "preferred_name"),
    )


def granola_plugin() -> PluginSpec:
    return PluginSpec(
        name="granola",
        description="read personal Granola notes through the local desktop session",
        runner=_runner("gotta.plugins.granola"),
        route_target=_module_attr("gotta.plugins.granola", "route_target"),
        route_priority=65,
        canonical_locator=_module_attr("gotta.plugins.granola", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.granola", "preferred_name"),
    )


def gsheets_plugin() -> PluginSpec:
    return PluginSpec(
        name="gsheets",
        description="read and search Google Sheets through the Sheets/Drive APIs",
        runner=_runner("gotta.plugins.gsheets"),
        route_target=_module_attr("gotta.plugins.gsheets", "route_target"),
        route_priority=55,
        canonical_locator=_module_attr("gotta.plugins.gsheets", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.gsheets", "preferred_name"),
    )


def gdrive_plugin() -> PluginSpec:
    return PluginSpec(
        name="gdrive",
        description="inspect and fetch Google Drive files through the Drive API",
        runner=_runner("gotta.plugins.gdrive"),
        route_target=_module_attr("gotta.plugins.gdrive", "route_target"),
        route_priority=60,
        canonical_locator=_module_attr("gotta.plugins.gdrive", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.gdrive", "preferred_name"),
    )


def github_plugin() -> PluginSpec:
    return PluginSpec(
        name="github",
        description="render common GitHub URLs through the GitHub CLI",
        runner=_runner("gotta.plugins.github"),
        route_target=_module_attr("gotta.plugins.github", "route_target"),
        route_priority=10,
        canonical_locator=_module_attr("gotta.plugins.github", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.github", "preferred_name"),
    )


def jira_plugin() -> PluginSpec:
    return PluginSpec(
        name="jira",
        description="discover, read, and author Jira issues through Atlassian OAuth",
        runner=_runner("gotta.plugins.jira"),
        route_target=_module_attr("gotta.plugins.jira", "route_target"),
        route_priority=30,
        canonical_locator=_module_attr("gotta.plugins.jira", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.jira", "preferred_name"),
    )


def session_plugin() -> PluginSpec:
    return PluginSpec(
        name="session",
        description="inspect the active session-rooted content context",
        runner=_runner("gotta.plugins.session"),
        should_materialize=lambda argv: False,
    )


def slack_plugin() -> PluginSpec:
    return PluginSpec(
        name="slack",
        description="query the bounded local Slack archive and explicit sync surface",
        runner=_runner("gotta.plugins.slack"),
        route_target=_module_attr("gotta.plugins.slack", "route_target"),
        route_priority=40,
        canonical_locator=_module_attr("gotta.plugins.slack", "canonical_locator"),
        preferred_name=_module_attr("gotta.plugins.slack", "preferred_name"),
    )


def _builtin_core_plugins() -> dict[str, PluginSpec]:
    factories = (
        ask_plugin,
        confluence_plugin,
        gdocs_plugin,
        gdrive_plugin,
        goal_plugin,
        granola_plugin,
        github_plugin,
        gsheets_plugin,
        jira_plugin,
        logs_plugin,
        notes_plugin,
        oops_plugin,
        peer_plugin,
        read_plugin,
        session_plugin,
        slack_plugin,
        todo_plugin,
        want_plugin,
    )
    return {plugin.name: plugin for plugin in (factory() for factory in factories)}



"""Shared dispatcher and runtime helpers for packaged gotta commands."""

from __future__ import annotations

from collections.abc import Callable
from contextlib import contextmanager
import io
import json
import os
import re
import sys
from typing import Any

from gotta.content import (
    CONTENT_ENV,
    ContentError,
    CommonOptions,
    Materialization,
    ResolvedDirs,
    SESSION_ENV,
    artifact_locator,
    content_locator,
    materialize_bytes,
    resolve_dirs,
    session_is_initialized,
)
from gotta.builtin import (
    PluginSpec,
    available_plugins as discovered_plugin_names,
    get_plugin,
)
from gotta.invocation import (
    ResolvedInvocation,
    SUPPRESS_MATERIALIZATION_ENV as INVOCATION_SUPPRESS_MATERIALIZATION_ENV,
    canonical_locator as resolve_canonical_locator,
    infer_content_type as resolve_content_type,
    invocation_locator as resolve_invocation_locator,
    preferred_name as resolve_preferred_name,
    resolve_invocation,
    should_materialize as resolve_should_materialize,
)
from gotta.source import (
    derive_source_metadata_from_payload,
    normalize_source_timestamp,
    slack_timestamp_to_iso,
)


def die(message: str, code: int = 2) -> int:
    print(message, file=sys.stderr)
    return code


SUPPRESS_MATERIALIZATION_ENV = INVOCATION_SUPPRESS_MATERIALIZATION_ENV


def available_plugins() -> list[str]:
    return discovered_plugin_names()


def plugin_spec(plugin: str) -> PluginSpec | None:
    return get_plugin(plugin)


def print_usage() -> int:
    print("usage: gotta <plugin> [args...]", file=sys.stderr)
    print("", file=sys.stderr)
    print("canonical session-binding path: `gotta ...`", file=sys.stderr)
    print("", file=sys.stderr)
    print(
        "session investigative surfaces live under `gotta session`: "
        "`manifest`, `timeline`, `graph`, `leads`, `analyze`",
        file=sys.stderr,
    )
    print("", file=sys.stderr)
    print("use `gotta --help-all` for recursive command help", file=sys.stderr)
    print("", file=sys.stderr)
    print("available plugins:", file=sys.stderr)
    for plugin in available_plugins():
        spec = plugin_spec(plugin)
        description = spec.description if spec else ""
        if description:
            print(f"  - {plugin:<10} {description}", file=sys.stderr)
            continue
        print(f"  - {plugin}", file=sys.stderr)
    return 0


def load_plugin_runner(plugin: str) -> Callable[[list[str]], int]:
    spec = plugin_spec(plugin)
    if spec is None:
        raise KeyError(plugin)
    return spec.runner


def split_common_options(argv: list[str]) -> tuple[CommonOptions, list[str]]:
    session_dir: str | None = None
    content_dir: str | None = None
    save_as: str | None = None

    cleaned: list[str] = []
    index = 0
    while index < len(argv):
        item = argv[index]
        if item == "--":
            cleaned.extend(argv[index:])
            break
        if item.startswith("--session="):
            session_dir = item.split("=", 1)[1]
            index += 1
            continue
        if item == "--session":
            if index + 1 >= len(argv):
                raise ContentError("--session requires a value")
            session_dir = argv[index + 1]
            index += 2
            continue
        if item.startswith("--content-dir="):
            content_dir = item.split("=", 1)[1]
            index += 1
            continue
        if item == "--content-dir":
            if index + 1 >= len(argv):
                raise ContentError("--content-dir requires a value")
            content_dir = argv[index + 1]
            index += 2
            continue
        if item.startswith("--save-as="):
            save_as = item.split("=", 1)[1]
            index += 1
            continue
        if item == "--save-as":
            if index + 1 >= len(argv):
                raise ContentError("--save-as requires a value")
            save_as = argv[index + 1]
            index += 2
            continue
        cleaned.append(item)
        index += 1

    return CommonOptions(
        session_dir=session_dir,
        content_dir=content_dir,
        save_as=save_as,
    ), cleaned


def should_materialize(plugin: str, argv: list[str]) -> bool:
    return resolve_should_materialize(plugin, argv)


def invocation_locator(plugin: str, argv: list[str]) -> str:
    return resolve_invocation_locator(plugin, argv)


def canonical_locator(plugin: str, argv: list[str]) -> str:
    return resolve_canonical_locator(plugin, argv)


def derive_preferred_name(plugin: str, argv: list[str], options: CommonOptions) -> str:
    return resolve_preferred_name(plugin, argv, options)


def infer_content_type(plugin: str, argv: list[str], name: str) -> str:
    return resolve_content_type(plugin, argv, name)


class CapturedStdout(io.TextIOBase):
    """Capture text and binary writes through the standard stdout surface."""

    def __init__(self, encoding: str) -> None:
        self._encoding = encoding
        self._buffer = io.BytesIO()
        self.buffer = self._buffer

    @property
    def encoding(self) -> str:
        return self._encoding

    def write(self, data: str) -> int:
        payload = data.encode(self._encoding)
        self._buffer.write(payload)
        return len(data)

    def flush(self) -> None:
        return None

    def isatty(self) -> bool:
        return False

    def getvalue(self) -> bytes:
        return self._buffer.getvalue()


@contextmanager
def capture_stdout() -> Any:
    original_stdout = sys.stdout
    capture = CapturedStdout(getattr(original_stdout, "encoding", None) or "utf-8")
    sys.stdout = capture
    try:
        yield capture
    finally:
        sys.stdout = original_stdout


@contextmanager
def scoped_runtime_env(dirs: ResolvedDirs) -> Any:
    previous = {
        SESSION_ENV: os.environ.get(SESSION_ENV),
        CONTENT_ENV: os.environ.get(CONTENT_ENV),
    }
    os.environ[SESSION_ENV] = str(dirs.session_dir)
    os.environ[CONTENT_ENV] = str(dirs.content_dir)
    try:
        yield
    finally:
        for key, value in previous.items():
            if value is None:
                os.environ.pop(key, None)
            else:
                os.environ[key] = value


def _emit_captured(data: bytes) -> None:
    if not data:
        return
    sys.stdout.buffer.write(data)
    sys.stdout.flush()


def _json_value(data: bytes) -> Any | None:
    try:
        payload = json.loads(data.decode("utf-8"))
    except (UnicodeDecodeError, json.JSONDecodeError):
        return None
    return payload


_MARKDOWN_SOURCE_PATTERNS: tuple[tuple[re.Pattern[str], str], ...] = (
    (
        re.compile(r"^\s*-\s*(?:\*\*)?Created:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE),
        "source_created_at",
    ),
    (
        re.compile(r"^\s*-\s*(?:\*\*)?Updated:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE),
        "source_updated_at",
    ),
    (
        re.compile(r"^\s*-\s*(?:\*\*)?Modified:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE),
        "source_updated_at",
    ),
    (
        re.compile(r"^\s*-\s*(?:\*\*)?Published:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE),
        "source_published_at",
    ),
    (
        re.compile(r"^\s*-\s*(?:\*\*)?Authored:(?:\*\*)?\s*(?P<value>\S.+?)\s*$", re.MULTILINE),
        "source_created_at",
    ),
)


def _json_nested(payload: dict[str, Any], *path: str) -> str:
    current: Any = payload
    for key in path:
        if not isinstance(current, dict):
            return ""
        current = current.get(key)
    return str(current or "").strip()


def _extract_source_metadata_from_json(payload: Any) -> dict[str, str]:
    metadata = derive_source_metadata_from_payload(payload)
    if not isinstance(payload, dict):
        return metadata
    candidates = (
        ("source_published_at", _json_nested(payload, "published_at")),
        ("source_published_at", _json_nested(payload, "created_at")),
        ("source_updated_at", _json_nested(payload, "updatedAt")),
        ("source_created_at", _json_nested(payload, "createdAt")),
        ("source_updated_at", _json_nested(payload, "updated")),
        ("source_created_at", _json_nested(payload, "created")),
        ("source_updated_at", _json_nested(payload, "modifiedTime")),
        ("source_created_at", _json_nested(payload, "createdTime")),
        ("source_created_at", _json_nested(payload, "commit", "author", "date")),
        ("source_created_at", _json_nested(payload, "author", "date")),
    )
    for key, value in candidates:
        parsed = normalize_source_timestamp(value) or str(value or "").strip()
        if parsed and key not in metadata:
            metadata[key] = parsed
    return metadata


def _extract_source_metadata_from_markdown(data: bytes) -> dict[str, str]:
    try:
        text = data.decode("utf-8")
    except UnicodeDecodeError:
        return {}
    metadata: dict[str, str] = {}
    for pattern, key in _MARKDOWN_SOURCE_PATTERNS:
        match = pattern.search(text)
        if not match:
            continue
        value = match.group("value").strip()
        parsed = normalize_source_timestamp(value) or value
        if parsed and key not in metadata:
            metadata[key] = parsed
    authored_match = re.search(r"\bauthored (?P<value>\d{4}-\d{2}-\d{2}T\S+Z?)\b", text)
    if authored_match and "source_created_at" not in metadata:
        metadata["source_created_at"] = authored_match.group("value")
    return metadata


def _derived_source_metadata(plugin: str, argv: list[str], data: bytes) -> dict[str, str]:
    metadata: dict[str, str] = {}
    canonical = canonical_locator(plugin, argv)
    defaults: dict[str, str] = {}
    if canonical.startswith("slack:thread:"):
        thread_ts = canonical.rsplit(":", 1)[-1]
        source_time = slack_timestamp_to_iso(thread_ts)
        if source_time:
            defaults["source_created_at"] = source_time
            defaults["source_updated_at"] = source_time
    if plugin != "slack" or not argv or argv[0] != "get":
        payload = _json_value(data)
        if payload is not None:
            metadata.update(_extract_source_metadata_from_json(payload))
        metadata.update(
            {
                key: value
                for key, value in _extract_source_metadata_from_markdown(data).items()
                if key not in metadata
            }
        )
        for key, value in defaults.items():
            metadata.setdefault(key, value)
        return metadata
    payload = _json_value(data)
    if isinstance(payload, dict):
        first_ts = slack_timestamp_to_iso(str(payload.get("firstTs") or ""))
        last_ts = slack_timestamp_to_iso(str(payload.get("lastTs") or ""))
        if first_ts:
            metadata["source_created_at"] = first_ts
        if last_ts:
            metadata["source_updated_at"] = last_ts
    metadata.update(
        {
            key: value
            for key, value in _extract_source_metadata_from_markdown(data).items()
            if key not in metadata
        }
    )
    for key, value in defaults.items():
        metadata.setdefault(key, value)
    return metadata


def _materialize_invocation(
    resolved_or_plugin: ResolvedInvocation | str,
    argv_or_data: list[str] | bytes,
    options: CommonOptions | None = None,
    data: bytes | None = None,
    *,
    dirs: ResolvedDirs,
) -> Materialization | None:
    if isinstance(resolved_or_plugin, ResolvedInvocation):
        resolved = resolved_or_plugin
        payload = argv_or_data if isinstance(argv_or_data, bytes) else b""
    else:
        plugin = resolved_or_plugin
        argv = list(argv_or_data) if isinstance(argv_or_data, list) else []
        payload = data or b""
        resolved = resolve_invocation(plugin, argv, options or CommonOptions())
    if not payload:
        return None
    materialize_plugin = resolved.resolved_plugin
    materialize_argv = resolved.resolved_argv
    actor = os.environ.get("GOTTA_PEER_LABEL", "").strip() or "primary"
    metadata = {
        "tool": "gotta",
        "plugin": materialize_plugin,
        "subcommand": materialize_argv[0] if materialize_argv else "",
        "argv": materialize_argv,
        "locator": invocation_locator(materialize_plugin, materialize_argv),
        "canonical_locator": resolved.canonical_locator,
        "source_kind": "stdin" if resolved.entry_plugin == "read" and resolved.entry_argv == ["-"] else "render",
        "content_type": resolved.content_type,
        "session_dir": str(dirs.session_dir),
        "content_dir": str(dirs.content_dir),
        "actor": actor,
    }
    if resolved.entry_plugin != materialize_plugin:
        metadata["entrypoint"] = resolved.entry_plugin
        metadata["entry_argv"] = resolved.entry_argv
        metadata["entry_locator"] = invocation_locator(resolved.entry_plugin, resolved.entry_argv)
        metadata["provider"] = resolved.provider
    peer_dir = os.environ.get("GOTTA_PEER_DIR", "").strip()
    if peer_dir:
        metadata["peer_dir"] = peer_dir
    invocation_id = os.environ.get("GOTTA_INVOCATION_ID", "").strip()
    if invocation_id:
        metadata["invocation_id"] = invocation_id
    metadata.update(_derived_source_metadata(materialize_plugin, materialize_argv, payload))
    return materialize_bytes(
        payload,
        dirs=dirs,
        preferred_name=resolved.preferred_name,
        metadata=metadata,
    )


def _emit_materialization_receipt(result: Materialization | None) -> None:
    if result is None:
        return
    print(
        f"stored content: {result.name_link} (data: {result.data_path})",
        file=sys.stderr,
    )
    print(
        "locators: "
        f"{artifact_locator(result.name_link.name, result.digest)}, "
        f"{content_locator(result.digest)}",
        file=sys.stderr,
    )


def _run_callable(func: Callable[[list[str]], int], argv: list[str]) -> int:
    return int(func(argv))


def require_operational_session(dirs: ResolvedDirs) -> None:
    if not session_is_initialized(dirs.session_dir):
        raise ContentError(
            "start or bind a session first with `gotta ...` or bootstrap one "
            "manually with `gotta session init --session <root>` before running "
            "operational commands"
        )


def run_plugin(plugin: str, argv: list[str]) -> int:
    if plugin in {"session", "work"}:
        options = CommonOptions()
        cleaned = argv
    else:
        try:
            options, cleaned = split_common_options(argv)
        except ContentError as exc:
            return die(str(exc))

    try:
        runner = load_plugin_runner(plugin)
    except KeyError:
        plugins = ", ".join(available_plugins())
        return die(f"unknown gotta plugin: {plugin}. available plugins: {plugins}")
    except RuntimeError as exc:
        return die(str(exc), code=1)

    resolved = resolve_invocation(plugin, cleaned, options)
    if not resolved.should_materialize:
        if options.session_dir or options.content_dir:
            try:
                dirs = resolve_dirs(options, create=False)
                require_operational_session(dirs)
            except ContentError as exc:
                return die(str(exc))
            with scoped_runtime_env(dirs):
                return _run_callable(runner, cleaned)
        return _run_callable(runner, cleaned)

    try:
        dirs = resolve_dirs(options, create=False)
        require_operational_session(dirs)
    except ContentError as exc:
        return die(str(exc))

    with scoped_runtime_env(dirs):
        with capture_stdout() as capture:
            code = _run_callable(runner, cleaned)
    data = capture.getvalue()
    if code == 0:
        try:
            result = _materialize_invocation(resolved, data, dirs=dirs)
        except ContentError as exc:
            return die(str(exc), code=1)
        _emit_materialization_receipt(result)
    _emit_captured(data)
    return code
